#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import dataclasses


@dataclasses.dataclass
class MultipartConfig:
    min_chunk_size: int
    max_chunk_size: int
    max_chunk_count: int
    max_single_part_size: int


class ClientConfig(object):
    def __init__(
        self,
        api_url,
        display_url,
        min_recommended_version,
        min_compatible_version,
        max_compatible_version,
        multipart_config,
    ):
        self._api_url = api_url
        self._display_url = display_url
        self._min_recommended_version = min_recommended_version
        self._min_compatible_version = min_compatible_version
        self._max_compatible_version = max_compatible_version
        self._multipart_config = multipart_config

    @property
    def api_url(self):
        return self._api_url

    @property
    def display_url(self):
        return self._display_url

    @property
    def min_recommended_version(self):
        return self._min_recommended_version

    @property
    def min_compatible_version(self):
        return self._min_compatible_version

    @property
    def max_compatible_version(self):
        return self._max_compatible_version

    @property
    def multipart_config(self):
        return self._multipart_config



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
from functools import wraps
from http.client import (
    NOT_FOUND,
    UNPROCESSABLE_ENTITY,
)

from requests.exceptions import HTTPError

from neptune.legacy.api_exceptions import (
    ExperimentNotFound,
    StorageLimitReached,
)
from neptune.legacy.exceptions import NeptuneException

_logger = logging.getLogger(__name__)


def extract_response_field(response, field_name):
    if response is None:
        return None

    try:
        response_json = response.json()
        if isinstance(response_json, dict):
            return response_json.get(field_name)
        else:
            _logger.debug("HTTP response is not a dict: %s", str(response_json))
            return None
    except ValueError as e:
        _logger.debug("Failed to parse HTTP response: %s", e)
        return None


def handle_quota_limits(f):
    """Wrapper for functions which may request for non existing experiment or cause quota limit breach

    Limitations:
    Decorated function must be called with experiment argument like this fun(..., experiment=<experiment>, ...)"""

    @wraps(f)
    def handler(*args, **kwargs):
        experiment = kwargs.get("experiment")
        if experiment is None:
            raise NeptuneException(
                "This function must be called with experiment passed by name,"
                " like this fun(..., experiment=<experiment>, ...)"
            )
        try:
            return f(*args, **kwargs)
        except HTTPError as e:
            if e.response.status_code == NOT_FOUND:
                raise ExperimentNotFound(
                    experiment_short_id=experiment.id,
                    project_qualified_name=experiment._project.full_id,
                )
            if e.response.status_code == UNPROCESSABLE_ENTITY and extract_response_field(
                e.response, "title"
            ).startswith("Storage limit reached in organization: "):
                raise StorageLimitReached()
            raise

    return handler



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from neptune.common.exceptions import STYLES
from neptune.legacy import envs
from neptune.legacy.exceptions import NeptuneException


class NeptuneApiException(NeptuneException):
    pass


class NeptuneSSLVerificationError(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneSSLVerificationError-----------------------------------------------------------------------
{end}

Neptune client was unable to verify your SSL Certificate.

{bold}What could go wrong?{end}
    - You are behind a proxy that inspects traffic to Neptune servers.
        - Contact your network administrator
    - Your on-prem installation's SSL/TLS certificate is not recognized due to a custom Certificate Authority (CA).
        - To check run the following command in terminal:
            {bash}curl https://<your_domain>/api/backend/echo {end}
        - Where <your_domain> is the address that you use to access Neptune UI i.e. abc.com
        - Contact your network administrator if you get the following output:
            {fail}"curl: (60) server certificate verification failed..."{end}
    - Your machine software is not up-to-date.
        - Minimal OS requirements:
            - Windows >= XP SP3
            - macOS >= 10.12.1
            - Ubuntu >= 12.04
            - Debian >= 8

{bold}What can I do?{end}
You can manually configure Neptune to skip all SSL checks. To do that
set the NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE environment variable to 'TRUE'.
{bold}Note that might mean your connection is less secure{end}.

Linux/Unix
In your terminal run:
    {bash}export NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE='TRUE'{end}

Windows
In your terminal run:
    {bash}set NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE='TRUE'{end}

Jupyter notebook
In your code cell:
    {bash}%env NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE='TRUE'{end}

You may also want to check the following docs pages:
    - https://docs.neptune.ai/api/environment_variables/#neptune_allow_self_signed_certificate


{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(**STYLES))


class ConnectionLost(NeptuneApiException):
    def __init__(self):
        super(ConnectionLost, self).__init__("Connection lost. Please try again.")


class ServerError(NeptuneApiException):
    def __init__(self):
        message = """
{h1}
----ServerError-----------------------------------------------------------------------
{end}
Neptune Client Library encountered an unexpected Server Error.

Please try again later or contact Neptune support.
"""
        super(ServerError, self).__init__(message.format(**STYLES))


class Unauthorized(NeptuneApiException):
    def __init__(self):
        message = """
{h1}
----Unauthorized-----------------------------------------------------------------------
{end}
You have no permission to access given resource.

    - Verify your API token is correct.
      See: https://docs-legacy.neptune.ai/security-and-privacy/api-tokens/how-to-find-and-set-neptune-api-token.html

    - Verify if you set your Project qualified name correctly
      The correct project qualified name should look like this {correct}WORKSPACE/PROJECT_NAME{end}.
      It has two parts:
          - {correct}WORKSPACE{end}: which can be your username or your organization name
          - {correct}PROJECT_NAME{end}: which is the actual project name you chose

    - Ask your organization administrator to grant you necessary privileges to the project
"""
        super(Unauthorized, self).__init__(message.format(**STYLES))


class Forbidden(NeptuneApiException):
    def __init__(self):
        message = """
{h1}
----Forbidden-----------------------------------------------------------------------
{end}
You have no permission to access given resource.

    - Verify your API token is correct.
      See: https://docs-legacy.neptune.ai/security-and-privacy/api-tokens/how-to-find-and-set-neptune-api-token.html

    - Verify if you set your Project qualified name correctly
      The correct project qualified name should look like this {correct}WORKSPACE/PROJECT_NAME{end}.
      It has two parts:
          - {correct}WORKSPACE{end}: which can be your username or your organization name
          - {correct}PROJECT_NAME{end}: which is the actual project name you chose

   - Ask your organization administrator to grant you necessary privileges to the project
"""
        super(Forbidden, self).__init__(message.format(**STYLES))


class InvalidApiKey(NeptuneApiException):
    def __init__(self):
        message = """
{h1}
----InvalidApiKey-----------------------------------------------------------------------
{end}
Your API token is invalid.

Learn how to get it in this docs page:
https://docs-legacy.neptune.ai/security-and-privacy/api-tokens/how-to-find-and-set-neptune-api-token.html

There are two options to add it:
    - specify it in your code
    - set an environment variable in your operating system.

{h2}CODE{end}
Pass the token to {bold}neptune.init(){end} via {bold}api_token{end} argument:
    {python}neptune.init(project_qualified_name='WORKSPACE_NAME/PROJECT_NAME', api_token='YOUR_API_TOKEN'){end}

{h2}ENVIRONMENT VARIABLE{end} {correct}(Recommended option){end}
or export or set an environment variable depending on your operating system:

    {correct}Linux/Unix{end}
    In your terminal run:
        {bash}export {env_api_token}=YOUR_API_TOKEN{end}

    {correct}Windows{end}
    In your CMD run:
        {bash}set {env_api_token}=YOUR_API_TOKEN{end}

and skip the {bold}api_token{end} argument of {bold}neptune.init(){end}:
    {python}neptune.init(project_qualified_name='WORKSPACE_NAME/PROJECT_NAME'){end}

You may also want to check the following docs pages:
    - https://docs-legacy.neptune.ai/security-and-privacy/api-tokens/how-to-find-and-set-neptune-api-token.html
    - https://docs-legacy.neptune.ai/getting-started/quick-starts/log_first_experiment.html

{correct}Need help?{end}-> https://docs-legacy.neptune.ai/getting-started/getting-help.html
"""
        super(InvalidApiKey, self).__init__(message.format(env_api_token=envs.API_TOKEN_ENV_NAME, **STYLES))


class WorkspaceNotFound(NeptuneApiException):
    def __init__(self, namespace_name):
        message = """
{h1}
----WorkspaceNotFound-------------------------------------------------------------------------
{end}
Workspace {python}{workspace}{end} not found.

Workspace is your username or a name of your team organization.
"""
        super(WorkspaceNotFound, self).__init__(message.format(workspace=namespace_name, **STYLES))


class ProjectNotFound(NeptuneApiException):
    def __init__(self, project_identifier):
        message = """
{h1}
----ProjectNotFound-------------------------------------------------------------------------
{end}
Project {python}{project}{end} not found.

Verify if your project's name was not misspelled. You can find proper name after logging into Neptune UI.
"""
        super(ProjectNotFound, self).__init__(message.format(project=project_identifier, **STYLES))


class PathInProjectNotFound(NeptuneApiException):
    def __init__(self, path, project_identifier):
        super(PathInProjectNotFound, self).__init__(
            "Path {} was not found in project {}.".format(path, project_identifier)
        )


class PathInExperimentNotFound(NeptuneApiException):
    def __init__(self, path, exp_identifier):
        super().__init__(f"Path {path} was not found in experiment {exp_identifier}.")


class NotebookNotFound(NeptuneApiException):
    def __init__(self, notebook_id, project=None):
        if project:
            super(NotebookNotFound, self).__init__(
                "Notebook '{}' not found in project '{}'.".format(notebook_id, project)
            )
        else:
            super(NotebookNotFound, self).__init__("Notebook '{}' not found.".format(notebook_id))


class ExperimentNotFound(NeptuneApiException):
    def __init__(self, experiment_short_id, project_qualified_name):
        super(ExperimentNotFound, self).__init__(
            "Experiment '{exp}' not found in '{project}'.".format(
                exp=experiment_short_id, project=project_qualified_name
            )
        )


class ChannelNotFound(NeptuneApiException):
    def __init__(self, channel_id):
        super(ChannelNotFound, self).__init__("Channel '{id}' not found.".format(id=channel_id))


class ExperimentAlreadyFinished(NeptuneApiException):
    def __init__(self, experiment_short_id):
        super(ExperimentAlreadyFinished, self).__init__(
            "Experiment '{}' is already finished.".format(experiment_short_id)
        )


class ExperimentLimitReached(NeptuneApiException):
    def __init__(self):
        super(ExperimentLimitReached, self).__init__("Experiment limit reached.")


class StorageLimitReached(NeptuneApiException):
    def __init__(self):
        super(StorageLimitReached, self).__init__("Storage limit reached.")


class ExperimentValidationError(NeptuneApiException):
    pass


class ChannelAlreadyExists(NeptuneApiException):
    def __init__(self, experiment_short_id, channel_name):
        super(ChannelAlreadyExists, self).__init__(
            "Channel with name '{}' already exists in experiment '{}'.".format(channel_name, experiment_short_id)
        )


class ChannelDoesNotExist(NeptuneApiException):
    def __init__(self, experiment_short_id, channel_name):
        super(ChannelDoesNotExist, self).__init__(
            "Channel with name '{}' does not exist in experiment '{}'.".format(channel_name, experiment_short_id)
        )


class ChannelsValuesSendBatchError(NeptuneApiException):
    @staticmethod
    def _format_error(error):
        return "{msg} (metricId: '{channelId}', x: {x})".format(msg=error.error, channelId=error.channelId, x=error.x)

    def __init__(self, experiment_short_id, batch_errors):
        super(ChannelsValuesSendBatchError, self).__init__(
            "Received batch errors sending channels' values to experiment {}. "
            "Cause: {} "
            "Skipping {} values.".format(
                experiment_short_id,
                self._format_error(batch_errors[0]) if batch_errors else "No errors",
                len(batch_errors),
            )
        )


class ExperimentOperationErrors(NeptuneApiException):
    """Handles minor errors returned by calling `client.executeOperations`"""

    def __init__(self, errors):
        super().__init__()
        self.errors = errors

    def __str__(self):
        lines = ["Caused by:"]
        for error in self.errors:
            lines.append(f"\t* {error}")
        return "\n".join(lines)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "replace_patch_version",
    "verify_type",
    "is_stream",
    "is_bool",
    "is_int",
    "is_float",
    "is_string",
    "is_float_like",
    "is_dict_like",
    "is_string_like",
    "is_stringify_value",
    "verify_collection_type",
    "verify_optional_callable",
    "is_collection",
    "base64_encode",
    "base64_decode",
    "get_absolute_paths",
    "get_common_root",
    "does_paths_share_common_drive",
    "is_ipython",
    "as_list",
]

import base64
import logging
import os
from glob import glob
from io import IOBase
from typing import (
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    TypeVar,
    Union,
)

from neptune.internal.types.stringify_value import StringifyValue

T = TypeVar("T")

_logger = logging.getLogger(__name__)


def replace_patch_version(version: str):
    return version[: version.index(".", version.index(".") + 1)] + ".0"


def verify_type(var_name: str, var, expected_type: Union[type, tuple]):
    try:
        if isinstance(expected_type, tuple):
            type_name = " or ".join(get_type_name(t) for t in expected_type)
        else:
            type_name = get_type_name(expected_type)
    except Exception as e:
        # Just to be sure that nothing weird will be raised here
        raise TypeError("Incorrect type of {}".format(var_name)) from e

    if not isinstance(var, expected_type):
        raise TypeError("{} must be a {} (was {})".format(var_name, type_name, type(var)))

    if isinstance(var, IOBase) and not hasattr(var, "read"):
        raise TypeError("{} is a stream, which does not implement read method".format(var_name))


def is_stream(var):
    return isinstance(var, IOBase) and hasattr(var, "read")


def is_bool(var):
    return isinstance(var, bool)


def is_int(var):
    return isinstance(var, int)


def is_float(var):
    return isinstance(var, (float, int))


def is_string(var):
    return isinstance(var, str)


def is_float_like(var):
    try:
        _ = float(var)
        return True
    except (ValueError, TypeError):
        return False


def is_dict_like(var):
    return isinstance(var, (dict, Mapping))


def is_string_like(var):
    try:
        _ = str(var)
        return True
    except ValueError:
        return False


def is_stringify_value(var):
    return isinstance(var, StringifyValue)


def get_type_name(_type: Union[type, tuple]):
    return _type.__name__ if hasattr(_type, "__name__") else str(_type)


def verify_collection_type(var_name: str, var, expected_type: Union[type, tuple]):
    verify_type(var_name, var, (list, set, tuple))
    for value in var:
        verify_type("elements of collection '{}'".format(var_name), value, expected_type)


def verify_optional_callable(var_name: str, var):
    if var and not callable(var):
        raise TypeError("{} must be a callable (was {})".format(var_name, type(var)))


def is_collection(var) -> bool:
    return isinstance(var, (list, set, tuple))


def base64_encode(data: bytes) -> str:
    return base64.b64encode(data).decode("utf-8")


def base64_decode(data: str) -> bytes:
    return base64.b64decode(data.encode("utf-8"))


def get_absolute_paths(file_globs: Iterable[str]) -> List[str]:
    expanded_paths: Set[str] = set()
    for file_glob in file_globs:
        expanded_paths |= set(glob(file_glob, recursive=True))
    return list(os.path.abspath(expanded_file) for expanded_file in expanded_paths)


def get_common_root(absolute_paths: List[str]) -> Optional[str]:
    try:
        common_root = os.path.commonpath(absolute_paths)
        if os.path.isfile(common_root):
            common_root = os.path.dirname(common_root)
        if common_root.startswith(os.getcwd() + os.sep):
            common_root = os.getcwd()
        return common_root
    except ValueError:
        return None


def does_paths_share_common_drive(paths: List[str]) -> bool:
    return len(set(map(lambda path: os.path.splitdrive(path)[0], paths))) == 1


def is_ipython() -> bool:
    try:
        import IPython

        ipython = IPython.core.getipython.get_ipython()
        return ipython is not None
    except ImportError:
        return False


def as_list(name: str, value: Optional[Union[str, Iterable[str]]]) -> Optional[Iterable[str]]:
    verify_type(name, value, (type(None), str, Iterable))
    if value is None:
        return []
    if isinstance(value, str):
        return [value]
    verify_collection_type(name, value, str)
    return value



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import time

import requests
from bravado.exception import (
    BravadoConnectionError,
    BravadoTimeoutError,
    HTTPBadGateway,
    HTTPForbidden,
    HTTPGatewayTimeout,
    HTTPInternalServerError,
    HTTPRequestTimeout,
    HTTPServiceUnavailable,
    HTTPTooManyRequests,
    HTTPUnauthorized,
)
from urllib3.exceptions import NewConnectionError

from neptune.legacy.api_exceptions import (
    ConnectionLost,
    Forbidden,
    NeptuneSSLVerificationError,
    ServerError,
    Unauthorized,
)

_logger = logging.getLogger(__name__)


def legacy_with_api_exceptions_handler(func):
    def wrapper(*args, **kwargs):
        retries = 11
        retry = 0
        while retry < retries:
            try:
                return func(*args, **kwargs)
            except requests.exceptions.SSLError:
                raise NeptuneSSLVerificationError()
            except HTTPServiceUnavailable:
                if retry >= 6:
                    _logger.warning("Experiencing connection interruptions. Reestablishing communication with Neptune.")
                time.sleep(2**retry)
                retry += 1
                continue
            except (
                BravadoConnectionError,
                BravadoTimeoutError,
                requests.exceptions.ConnectionError,
                requests.exceptions.Timeout,
                HTTPRequestTimeout,
                HTTPGatewayTimeout,
                HTTPBadGateway,
                HTTPTooManyRequests,
                HTTPInternalServerError,
                NewConnectionError,
            ):
                if retry >= 6:
                    _logger.warning("Experiencing connection interruptions. Reestablishing communication with Neptune.")
                time.sleep(2**retry)
                retry += 1
                continue
            except HTTPUnauthorized:
                raise Unauthorized()
            except HTTPForbidden:
                raise Forbidden()
            except requests.exceptions.RequestException as e:
                if e.response is None:
                    raise
                status_code = e.response.status_code
                if status_code in (
                    HTTPRequestTimeout.status_code,
                    HTTPBadGateway.status_code,
                    HTTPServiceUnavailable.status_code,
                    HTTPGatewayTimeout.status_code,
                    HTTPTooManyRequests.status_code,
                    HTTPInternalServerError.status_code,
                ):
                    if retry >= 6:
                        _logger.warning(
                            "Experiencing connection interruptions. Reestablishing communication with Neptune."
                        )
                    time.sleep(2**retry)
                    retry += 1
                    continue
                elif status_code >= HTTPInternalServerError.status_code:
                    raise ServerError()
                elif status_code == HTTPUnauthorized.status_code:
                    raise Unauthorized()
                elif status_code == HTTPForbidden.status_code:
                    raise Forbidden()
                else:
                    raise
        raise ConnectionLost()

    return wrapper



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["project_name_lookup"]

import logging
import os
from typing import Optional

from neptune.envs import PROJECT_ENV_NAME
from neptune.exceptions import NeptuneMissingProjectNameException
from neptune.internal.backends.api_model import Project
from neptune.internal.backends.neptune_backend import NeptuneBackend
from neptune.internal.id_formats import QualifiedName
from neptune.internal.utils import verify_type

_logger = logging.getLogger(__name__)


def project_name_lookup(backend: NeptuneBackend, name: Optional[QualifiedName] = None) -> Project:
    verify_type("name", name, (str, type(None)))

    if not name:
        name = os.getenv(PROJECT_ENV_NAME)
    if not name:
        available_workspaces = backend.get_available_workspaces()
        available_projects = backend.get_available_projects()

        raise NeptuneMissingProjectNameException(
            available_workspaces=available_workspaces,
            available_projects=available_projects,
        )

    return backend.get_project(name)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["kill_me"]

import os
import signal

from neptune.envs import NEPTUNE_SUBPROCESS_KILL_TIMEOUT

try:
    import psutil

    PSUTIL_INSTALLED = True
except ImportError:
    PSUTIL_INSTALLED = False


KILL_TIMEOUT = int(os.getenv(NEPTUNE_SUBPROCESS_KILL_TIMEOUT, "5"))


def kill_me():
    if PSUTIL_INSTALLED:
        process = psutil.Process(os.getpid())
        try:
            children = _get_process_children(process)
        except psutil.NoSuchProcess:
            children = []

        for child_proc in children:
            _terminate(child_proc)
        _, alive = psutil.wait_procs(children, timeout=KILL_TIMEOUT)
        for child_proc in alive:
            _kill(child_proc)
        # finish with terminating self
        _terminate(process)
    else:
        os.kill(os.getpid(), signal.SIGINT)


def _terminate(process):
    try:
        process.terminate()
    except psutil.NoSuchProcess:
        pass


def _kill(process):
    try:
        if process.is_running():
            process.kill()
    except psutil.NoSuchProcess:
        pass


def _get_process_children(process):
    try:
        return process.children(recursive=True)
    except psutil.NoSuchProcess:
        return []



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["String"]

import typing

from neptune.attributes.atoms.copiable_atom import CopiableAtom
from neptune.internal.container_type import ContainerType
from neptune.internal.operation import AssignString
from neptune.internal.utils.logger import logger
from neptune.internal.utils.paths import path_to_str
from neptune.types.atoms.string import String as StringVal

if typing.TYPE_CHECKING:
    from neptune.internal.backends.neptune_backend import NeptuneBackend
    from neptune.metadata_containers import MetadataContainer


class String(CopiableAtom):

    MAX_VALUE_LENGTH = 16384

    def __init__(self, container: "MetadataContainer", path: typing.List[str]):
        super().__init__(container, path)
        self._value_truncation_occurred = False

    @staticmethod
    def create_assignment_operation(path, value: str):
        return AssignString(path, value)

    @staticmethod
    def getter(
        backend: "NeptuneBackend",
        container_id: str,
        container_type: ContainerType,
        path: typing.List[str],
    ) -> str:
        val = backend.get_string_attribute(container_id, container_type, path)
        return val.value

    def assign(self, value: typing.Union[StringVal, str], *, wait: bool = False):
        if not isinstance(value, StringVal):
            value = StringVal(value)

        if len(value.value) > String.MAX_VALUE_LENGTH:
            value.value = value.value[: String.MAX_VALUE_LENGTH]

            if not self._value_truncation_occurred:
                # the first truncation
                self._value_truncation_occurred = True
                logger.warning(
                    "Warning: string '%s' value was"
                    " longer than %s characters and was truncated."
                    " This warning is printed only once.",
                    path_to_str(self._path),
                    String.MAX_VALUE_LENGTH,
                )

        with self._container.lock():
            self._enqueue_operation(self.create_assignment_operation(self._path, value.value), wait=wait)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["FileSeries"]

import time
from itertools import cycle
from typing import (
    TYPE_CHECKING,
    List,
    Optional,
    Sequence,
    TypeVar,
)

from neptune.internal.types.stringify_value import extract_if_stringify_value
from neptune.internal.utils import is_collection
from neptune.internal.utils.logger import logger
from neptune.types import File
from neptune.types.series.series import Series

if TYPE_CHECKING:
    from neptune.types.value_visitor import ValueVisitor

Ret = TypeVar("Ret")


class FileSeries(Series):
    def __init__(
        self, values, timestamps: Optional[Sequence[float]] = None, steps: Optional[Sequence[float]] = None, **kwargs
    ):
        values = extract_if_stringify_value(values)

        if not is_collection(values):
            raise TypeError("`values` is not a collection")

        self._values = [File.create_from(value) for value in values]
        self.name = kwargs.pop("name", None)
        self.description = kwargs.pop("description", None)
        if kwargs:
            logger.error("Warning: unexpected arguments (%s) in FileSeries", kwargs)

        if steps is None:
            self._steps = cycle([None])
        else:
            assert len(values) == len(steps)
            self._steps = steps

        if timestamps is None:
            self._timestamps = cycle([time.time()])
        else:
            assert len(values) == len(timestamps)
            self._timestamps = timestamps

    @property
    def steps(self):
        return self._steps

    @property
    def timestamps(self):
        return self._timestamps

    def accept(self, visitor: "ValueVisitor[Ret]") -> Ret:
        return visitor.visit_image_series(self)

    @property
    def values(self) -> List[File]:
        return self._values

    def __str__(self):
        return f"FileSeries({self.values})"



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from neptune.common.hardware.constants import BYTES_IN_ONE_GB
from neptune.common.hardware.metrics.metric import (
    Metric,
    MetricResourceType,
)
from neptune.common.hardware.metrics.metrics_container import MetricsContainer


class MetricsFactory(object):
    def __init__(self, gauge_factory, system_resource_info):
        self.__gauge_factory = gauge_factory
        self.__system_resource_info = system_resource_info

    def create_metrics_container(self):
        cpu_usage_metric = self.__create_cpu_usage_metric()
        memory_metric = self.__create_memory_metric()

        has_gpu = self.__system_resource_info.has_gpu()
        gpu_usage_metric = self.__create_gpu_usage_metric() if has_gpu else None
        gpu_memory_metric = self.__create_gpu_memory_metric() if has_gpu else None

        return MetricsContainer(
            cpu_usage_metric=cpu_usage_metric,
            memory_metric=memory_metric,
            gpu_usage_metric=gpu_usage_metric,
            gpu_memory_metric=gpu_memory_metric,
        )

    def __create_cpu_usage_metric(self):
        return Metric(
            name="CPU - usage",
            description="average of all cores",
            resource_type=MetricResourceType.CPU,
            unit="%",
            min_value=0.0,
            max_value=100.0,
            gauges=[self.__gauge_factory.create_cpu_usage_gauge()],
        )

    def __create_memory_metric(self):
        return Metric(
            name="RAM",
            description="",
            resource_type=MetricResourceType.RAM,
            unit="GB",
            min_value=0.0,
            max_value=self.__system_resource_info.memory_amount_bytes / float(BYTES_IN_ONE_GB),
            gauges=[self.__gauge_factory.create_memory_usage_gauge()],
        )

    def __create_gpu_usage_metric(self):
        return Metric(
            name="GPU - usage",
            description="{} cards".format(self.__system_resource_info.gpu_card_count),
            resource_type=MetricResourceType.GPU,
            unit="%",
            min_value=0.0,
            max_value=100.0,
            gauges=[
                self.__gauge_factory.create_gpu_usage_gauge(card_index=card_index)
                for card_index in self.__system_resource_info.gpu_card_indices
            ],
        )

    def __create_gpu_memory_metric(self):
        return Metric(
            name="GPU - memory",
            description="{} cards".format(self.__system_resource_info.gpu_card_count),
            resource_type=MetricResourceType.GPU_RAM,
            unit="GB",
            min_value=0.0,
            max_value=self.__system_resource_info.gpu_memory_amount_bytes / float(BYTES_IN_ONE_GB),
            gauges=[
                self.__gauge_factory.create_gpu_memory_gauge(card_index=card_index)
                for card_index in self.__system_resource_info.gpu_card_indices
            ],
        )



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["MetadataContainer"]

import abc
import atexit
import itertools
import os
import threading
import time
import traceback
from contextlib import AbstractContextManager
from functools import (
    partial,
    wraps,
)
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Union,
)

from neptune.attributes import create_attribute_from_type
from neptune.attributes.attribute import Attribute
from neptune.attributes.namespace import Namespace as NamespaceAttr
from neptune.attributes.namespace import NamespaceBuilder
from neptune.common.exceptions import UNIX_STYLES
from neptune.common.utils import reset_internal_ssl_state
from neptune.common.warnings import warn_about_unsupported_type
from neptune.envs import (
    NEPTUNE_ENABLE_DEFAULT_ASYNC_LAG_CALLBACK,
    NEPTUNE_ENABLE_DEFAULT_ASYNC_NO_PROGRESS_CALLBACK,
)
from neptune.exceptions import (
    MetadataInconsistency,
    NeptunePossibleLegacyUsageException,
)
from neptune.handler import Handler
from neptune.internal.backends.api_model import (
    ApiExperiment,
    AttributeType,
    Project,
)
from neptune.internal.backends.factory import get_backend
from neptune.internal.backends.neptune_backend import NeptuneBackend
from neptune.internal.backends.nql import NQLQuery
from neptune.internal.backends.project_name_lookup import project_name_lookup
from neptune.internal.backgroud_job_list import BackgroundJobList
from neptune.internal.container_structure import ContainerStructure
from neptune.internal.container_type import ContainerType
from neptune.internal.id_formats import (
    QualifiedName,
    SysId,
    UniqueId,
    conform_optional,
)
from neptune.internal.init.parameters import (
    ASYNC_LAG_THRESHOLD,
    ASYNC_NO_PROGRESS_THRESHOLD,
    DEFAULT_FLUSH_PERIOD,
)
from neptune.internal.operation import DeleteAttribute
from neptune.internal.operation_processors.factory import get_operation_processor
from neptune.internal.operation_processors.operation_processor import OperationProcessor
from neptune.internal.state import ContainerState
from neptune.internal.utils import (
    verify_optional_callable,
    verify_type,
)
from neptune.internal.utils.logger import logger
from neptune.internal.utils.paths import parse_path
from neptune.internal.utils.uncaught_exception_handler import instance as uncaught_exception_handler
from neptune.internal.value_to_attribute_visitor import ValueToAttributeVisitor
from neptune.metadata_containers.abstract import (
    NeptuneObject,
    NeptuneObjectCallback,
)
from neptune.metadata_containers.metadata_containers_table import Table
from neptune.types.mode import Mode
from neptune.types.type_casting import cast_value
from neptune.utils import stop_synchronization_callback


def ensure_not_stopped(fun):
    @wraps(fun)
    def inner_fun(self: "MetadataContainer", *args, **kwargs):
        self._raise_if_stopped()
        return fun(self, *args, **kwargs)

    return inner_fun


class MetadataContainer(AbstractContextManager, NeptuneObject):
    container_type: ContainerType

    LEGACY_METHODS = set()

    def __init__(
        self,
        *,
        project: Optional[str] = None,
        api_token: Optional[str] = None,
        mode: Mode = Mode.ASYNC,
        flush_period: float = DEFAULT_FLUSH_PERIOD,
        proxies: Optional[dict] = None,
        async_lag_callback: Optional[NeptuneObjectCallback] = None,
        async_lag_threshold: float = ASYNC_LAG_THRESHOLD,
        async_no_progress_callback: Optional[NeptuneObjectCallback] = None,
        async_no_progress_threshold: float = ASYNC_NO_PROGRESS_THRESHOLD,
    ):
        verify_type("project", project, (str, type(None)))
        verify_type("api_token", api_token, (str, type(None)))
        verify_type("mode", mode, Mode)
        verify_type("flush_period", flush_period, (int, float))
        verify_type("proxies", proxies, (dict, type(None)))
        verify_type("async_lag_threshold", async_lag_threshold, (int, float))
        verify_optional_callable("async_lag_callback", async_lag_callback)
        verify_type("async_no_progress_threshold", async_no_progress_threshold, (int, float))
        verify_optional_callable("async_no_progress_callback", async_no_progress_callback)

        self._mode: Mode = mode
        self._flush_period = flush_period
        self._lock: threading.RLock = threading.RLock()
        self._forking_cond: threading.Condition = threading.Condition()
        self._forking_state: bool = False
        self._state: ContainerState = ContainerState.CREATED

        self._backend: NeptuneBackend = get_backend(mode=mode, api_token=api_token, proxies=proxies)

        self._project_qualified_name: Optional[str] = conform_optional(project, QualifiedName)
        self._project_api_object: Project = project_name_lookup(
            backend=self._backend, name=self._project_qualified_name
        )
        self._project_id: UniqueId = self._project_api_object.id

        self._api_object: ApiExperiment = self._get_or_create_api_object()
        self._id: UniqueId = self._api_object.id
        self._sys_id: SysId = self._api_object.sys_id
        self._workspace: str = self._api_object.workspace
        self._project_name: str = self._api_object.project_name

        self._async_lag_threshold = async_lag_threshold
        self._async_lag_callback = self._get_callback(
            provided=async_lag_callback,
            env_name=NEPTUNE_ENABLE_DEFAULT_ASYNC_LAG_CALLBACK,
        )
        self._async_no_progress_threshold = async_no_progress_threshold
        self._async_no_progress_callback = self._get_callback(
            provided=async_no_progress_callback,
            env_name=NEPTUNE_ENABLE_DEFAULT_ASYNC_NO_PROGRESS_CALLBACK,
        )

        self._op_processor: OperationProcessor = get_operation_processor(
            mode=mode,
            container_id=self._id,
            container_type=self.container_type,
            backend=self._backend,
            lock=self._lock,
            flush_period=flush_period,
            async_lag_callback=partial(self._async_lag_callback, self) if self._async_lag_callback else None,
            async_lag_threshold=self._async_lag_threshold,
            async_no_progress_callback=partial(self._async_no_progress_callback, self)
            if self._async_no_progress_callback
            else None,
            async_no_progress_threshold=self._async_no_progress_threshold,
        )
        self._bg_job: BackgroundJobList = self._prepare_background_jobs_if_non_read_only()
        self._structure: ContainerStructure[Attribute, NamespaceAttr] = ContainerStructure(NamespaceBuilder(self))

        if self._mode != Mode.OFFLINE:
            self.sync(wait=False)

        if self._mode != Mode.READ_ONLY:
            self._write_initial_attributes()

        self._startup(debug_mode=mode == Mode.DEBUG)

        try:
            os.register_at_fork(
                before=self._before_fork,
                after_in_child=self._handle_fork_in_child,
                after_in_parent=self._handle_fork_in_parent,
            )
        except AttributeError:
            pass

    """
    OpenSSL's internal random number generator does not properly handle forked processes.
    Applications must change the PRNG state of the parent process if they use any SSL feature with os.fork().
    Any successful call of RAND_add(), RAND_bytes() or RAND_pseudo_bytes() is sufficient.
    https://docs.python.org/3/library/ssl.html#multi-processing

    On Linux it looks like it does not help much but does not break anything either.
    """

    @staticmethod
    def _get_callback(provided: Optional[NeptuneObjectCallback], env_name: str) -> Optional[NeptuneObjectCallback]:
        if provided is not None:
            return provided
        if os.getenv(env_name, "") == "TRUE":
            return stop_synchronization_callback
        return None

    def _handle_fork_in_parent(self):
        reset_internal_ssl_state()
        if self._state == ContainerState.STARTED:
            self._op_processor.resume()
            self._bg_job.resume()

        with self._forking_cond:
            self._forking_state = False
            self._forking_cond.notify_all()

    def _handle_fork_in_child(self):
        reset_internal_ssl_state()
        if self._state == ContainerState.STARTED:
            self._op_processor.close()
            self._op_processor = get_operation_processor(
                mode=self._mode,
                container_id=self._id,
                container_type=self.container_type,
                backend=self._backend,
                lock=self._lock,
                flush_period=self._flush_period,
                async_lag_callback=partial(self._async_lag_callback, self) if self._async_lag_callback else None,
                async_lag_threshold=self._async_lag_threshold,
                async_no_progress_callback=partial(self._async_no_progress_callback, self)
                if self._async_no_progress_callback
                else None,
                async_no_progress_threshold=self._async_no_progress_threshold,
            )

            # TODO: Every implementation of background job should handle fork by itself.
            self._bg_job = BackgroundJobList([])

            self._op_processor.start()

        with self._forking_cond:
            self._forking_state = False
            self._forking_cond.notify_all()

    def _before_fork(self):
        with self._forking_cond:
            self._forking_cond.wait_for(lambda: self._state != ContainerState.STOPPING)
            self._forking_state = True

        if self._state == ContainerState.STARTED:
            self._bg_job.pause()
            self._op_processor.pause()

    def _prepare_background_jobs_if_non_read_only(self) -> BackgroundJobList:
        if self._mode != Mode.READ_ONLY:
            return self._prepare_background_jobs()
        return BackgroundJobList([])

    @abc.abstractmethod
    def _get_or_create_api_object(self) -> ApiExperiment:
        raise NotImplementedError

    def _prepare_background_jobs(self) -> BackgroundJobList:
        return BackgroundJobList([])

    def _write_initial_attributes(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_tb is not None:
            traceback.print_exception(exc_type, exc_val, exc_tb)
        self.stop()

    def __getattr__(self, item):
        if item in self.LEGACY_METHODS:
            raise NeptunePossibleLegacyUsageException()
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{item}'")

    @abc.abstractmethod
    def _raise_if_stopped(self):
        raise NotImplementedError

    def _get_subpath_suggestions(self, path_prefix: str = None, limit: int = 1000) -> List[str]:
        parsed_path = parse_path(path_prefix or "")
        return list(itertools.islice(self._structure.iterate_subpaths(parsed_path), limit))

    def _ipython_key_completions_(self):
        return self._get_subpath_suggestions()

    @ensure_not_stopped
    def __getitem__(self, path: str) -> "Handler":
        return Handler(self, path)

    @ensure_not_stopped
    def __setitem__(self, key: str, value) -> None:
        self.__getitem__(key).assign(value)

    @ensure_not_stopped
    def __delitem__(self, path) -> None:
        self.pop(path)

    @ensure_not_stopped
    def assign(self, value, *, wait: bool = False) -> None:
        """Assigns values to multiple fields from a dictionary.

        You can use this method to quickly log all parameters at once.

        Args:
            value (dict): A dictionary with values to assign, where keys become paths of the fields.
                The dictionary can be nested, in which case the path will be a combination of all the keys.
            wait: If `True`, Neptune waits to send all tracked metadata to the server before executing the call.

        Examples:
            >>> import neptune
            >>> run = neptune.init_run()
            >>> # Assign a single value with the Python "=" operator
            >>> run["parameters/learning_rate"] = 0.8
            >>> # or the assign() method
            >>> run["parameters/learning_rate"].assign(0.8)
            >>> # Assign a dictionary with the Python "=" operator
            >>> run["parameters"] = {"max_epochs": 10, "optimizer": "Adam", "learning_rate": 0.8}
            >>> # or the assign() method
            >>> run.assign({"parameters": {"max_epochs": 10, "optimizer": "Adam", "learning_rate": 0.8}})

            When operating on a handler object, you can use assign() to circumvent normal Python variable assignment.
            >>> params = run["params"]
            >>> params.assign({"max_epochs": 10, "optimizer": "Adam", "learning_rate": 0.8})

        See also the API reference:
            https://docs.neptune.ai/api/universal/#assign
        """
        self._get_root_handler().assign(value, wait=wait)

    @ensure_not_stopped
    def fetch(self) -> dict:
        """Fetch values of all non-File Atom fields as a dictionary.

        You can use this method to retrieve metadata from a started or resumed run.
        The result preserves the hierarchical structure of the run's metadata, but only contains Atom fields.
        This means fields that contain single values, as opposed to series, files, or sets.

        Returns:
            `dict` containing the values of all non-File Atom fields.

        Examples:
            Resuming an existing run and fetching metadata from it:
            >>> import neptune
            >>> resumed_run = neptune.init_run(with_id="CLS-3")
            >>> params = resumed_run["model/parameters"].fetch()
            >>> run_data = resumed_run.fetch()
            >>> print(run_data)
            >>> # prints all Atom attributes stored in run as a dict

            Fetching metadata from an existing model version:
            >>> model_version = neptune.init_model_version(with_id="CLS-TREE-45")
            >>> optimizer = model["parameters/optimizer"].fetch()

        See also the API reference:
            https://docs.neptune.ai/api/universal#fetch
        """
        return self._get_root_handler().fetch()

    def ping(self):
        self._backend.ping(self._id, self.container_type)

    def start(self):
        atexit.register(self._shutdown_hook)
        self._op_processor.start()
        self._bg_job.start(self)
        self._state = ContainerState.STARTED

    def stop(self, *, seconds: Optional[Union[float, int]] = None) -> None:
        """Stops the connection and ends the synchronization thread.

        You should stop any initialized runs or other objects when the connection to them is no longer needed.

        This method is automatically called:
        - when the script that created the run or other object finishes execution.
        - if using a context manager, on destruction of the Neptune context.

        Note: In interactive sessions, such as Jupyter Notebook, objects are stopped automatically only when
        the Python kernel stops. However, background monitoring of system metrics and standard streams is disabled
        unless explicitly enabled when initializing Neptune.

        Args:
            seconds: Seconds to wait for all metadata tracking calls to finish before stopping the object.
                If `None`, waits for all tracking calls to finish.

        Example:
            >>> import neptune
            >>> run = neptune.init_run()
            >>> # Your training or monitoring code
            >>> run.stop()

        See also the docs:
            Best practices - Stopping objects
                https://docs.neptune.ai/usage/best_practices/#stopping-runs-and-other-objects
            API reference:
                https://docs.neptune.ai/api/universal/#stop
        """
        verify_type("seconds", seconds, (float, int, type(None)))
        if self._state != ContainerState.STARTED:
            return

        with self._forking_cond:
            self._forking_cond.wait_for(lambda: not self._forking_state)
            self._state = ContainerState.STOPPING

        ts = time.time()
        logger.info("Shutting down background jobs, please wait a moment...")
        self._bg_job.stop()
        self._bg_job.join(seconds)
        logger.info("Done!")

        sec_left = None if seconds is None else seconds - (time.time() - ts)
        self._op_processor.stop(sec_left)

        if self._mode not in {Mode.OFFLINE, Mode.DEBUG}:
            logger.info("Explore the metadata in the Neptune app:")
            logger.info(self.get_url().rstrip("/") + "/metadata")
        self._backend.close()

        with self._forking_cond:
            self._state = ContainerState.STOPPED
            self._forking_cond.notify_all()

    def get_state(self) -> str:
        """Returns the current state of the container as a string.

        Examples:
            >>> from neptune import init_run
            >>> run = init_run()
            >>> run.get_state()
            'started'
            >>> run.stop()
            >>> run.get_state()
            'stopped'
        """
        return self._state.value

    def get_structure(self) -> Dict[str, Any]:
        """Returns the object's metadata structure as a dictionary.

        This method can be used to programmatically traverse the metadata structure of a run, model,
        or project object when using Neptune in automated workflows.

        Note: The returned object is a deep copy of the structure of the internal object.

        See also the API reference:
            https://docs.neptune.ai/api/universal/#get_structure
        """
        return self._structure.get_structure().to_dict()

    def print_structure(self) -> None:
        """Pretty-prints the structure of the object's metadata.

        Paths are ordered lexicographically and the whole structure is neatly colored.

        See also: https://docs.neptune.ai/api/universal/#print_structure
        """
        self._print_structure_impl(self.get_structure(), indent=0)

    def _print_structure_impl(self, struct: dict, indent: int) -> None:
        for key in sorted(struct.keys()):
            print("    " * indent, end="")
            if isinstance(struct[key], dict):
                print("{blue}'{key}'{end}:".format(blue=UNIX_STYLES["blue"], key=key, end=UNIX_STYLES["end"]))
                self._print_structure_impl(struct[key], indent=indent + 1)
            else:
                print(
                    "{blue}'{key}'{end}: {type}".format(
                        blue=UNIX_STYLES["blue"],
                        key=key,
                        end=UNIX_STYLES["end"],
                        type=type(struct[key]).__name__,
                    )
                )

    def define(
        self,
        path: str,
        value: Any,
        *,
        wait: bool = False,
    ) -> Optional[Attribute]:
        with self._lock:
            old_attr = self.get_attribute(path)
            if old_attr is not None:
                raise MetadataInconsistency("Attribute or namespace {} is already defined".format(path))

            neptune_value = cast_value(value)
            if neptune_value is None:
                warn_about_unsupported_type(type_str=str(type(value)))
                return None

            attr = ValueToAttributeVisitor(self, parse_path(path)).visit(neptune_value)
            self.set_attribute(path, attr)
            attr.process_assignment(neptune_value, wait=wait)
            return attr

    def get_attribute(self, path: str) -> Optional[Attribute]:
        with self._lock:
            return self._structure.get(parse_path(path))

    def set_attribute(self, path: str, attribute: Attribute) -> Optional[Attribute]:
        with self._lock:
            return self._structure.set(parse_path(path), attribute)

    def exists(self, path: str) -> bool:
        """Checks if there is a field or namespace under the specified path."""
        verify_type("path", path, str)
        return self.get_attribute(path) is not None

    @ensure_not_stopped
    def pop(self, path: str, *, wait: bool = False) -> None:
        """Removes the field stored under the path and all data associated with it.

        Args:
            path: Path of the field to be removed.
            wait: If `True`, Neptune waits to send all tracked metadata to the server before executing the call.

        Examples:
            >>> import neptune
            >>> run = neptune.init_run()
            >>> run["parameters/learninggg_rata"] = 0.3
            >>> # Let's delete that misspelled field along with its data
            ... run.pop("parameters/learninggg_rata")
            >>> run["parameters/learning_rate"] = 0.3
            >>> # Training finished
            ... run["trained_model"].upload("model.pt")
            >>> # "model_checkpoint" is a File field
            ... run.pop("model_checkpoint")

        See also the API reference:
           https://docs.neptune.ai/api/universal/#pop
        """
        verify_type("path", path, str)
        self._get_root_handler().pop(path, wait=wait)

    def _pop_impl(self, parsed_path: List[str], *, wait: bool):
        self._structure.pop(parsed_path)
        self._op_processor.enqueue_operation(DeleteAttribute(parsed_path), wait=wait)

    def lock(self) -> threading.RLock:
        return self._lock

    def wait(self, *, disk_only=False) -> None:
        """Wait for all the queued metadata tracking calls to reach the Neptune servers.

        Args:
            disk_only: If `True`, the process will only wait for data to be saved
                locally from memory, but will not wait for them to reach Neptune servers.

        See also the API reference:
            https://docs.neptune.ai/api/universal/#wait
        """
        with self._lock:
            if disk_only:
                self._op_processor.flush()
            else:
                self._op_processor.wait()

    def sync(self, *, wait: bool = True) -> None:
        """Synchronizes the local representation of the object with the representation on the Neptune servers.

        Args:
            wait: If `True`, the process will only wait for data to be saved
                locally from memory, but will not wait for them to reach Neptune servers.

        Example:
            >>> import neptune
            >>> # Connect to a run from Worker #3
            ... worker_id = 3
            >>> run = neptune.init_run(with_id="DIST-43", monitoring_namespace=f"monitoring/{worker_id}")
            >>> # Try to access logs that were created in the meantime by Worker #2
            ... worker_2_status = run["status/2"].fetch()
            ... # Error if this field was created after this script starts
            >>> run.sync() # Synchronizes local representation with Neptune servers
            >>> worker_2_status = run["status/2"].fetch()
            ... # No error

        See also the API reference:
            https://docs.neptune.ai/api/universal/#sync
        """
        with self._lock:
            if wait:
                self._op_processor.wait()
            attributes = self._backend.get_attributes(self._id, self.container_type)
            self._structure.clear()
            for attribute in attributes:
                self._define_attribute(parse_path(attribute.path), attribute.type)

    def _define_attribute(self, _path: List[str], _type: AttributeType):
        attr = create_attribute_from_type(_type, self, _path)
        self._structure.set(_path, attr)

    def _get_root_handler(self):
        return Handler(self, "")

    @abc.abstractmethod
    def get_url(self) -> str:
        """Returns a link to the object in the Neptune app.

        The same link is printed in the console once the object has been initialized.

        API reference: https://docs.neptune.ai/api/universal/#get_url
        """
        ...

    def _startup(self, debug_mode):
        if not debug_mode:
            logger.info(self.get_url())

        self.start()

        uncaught_exception_handler.activate()

    def _shutdown_hook(self):
        self.stop()

    def _fetch_entries(self, child_type: ContainerType, query: NQLQuery, columns: Optional[Iterable[str]]) -> Table:
        if columns is not None:
            # always return entries with `sys/id` column when filter applied
            columns = set(columns)
            columns.add("sys/id")

        leaderboard_entries = self._backend.search_leaderboard_entries(
            project_id=self._project_id,
            types=[child_type],
            query=query,
            columns=columns,
        )

        return Table(
            backend=self._backend,
            container_type=child_type,
            entries=leaderboard_entries,
        )

    def get_root_object(self) -> "MetadataContainer":
        """Returns the same Neptune object."""
        return self



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["custom_run_id_exceeds_length", "image_size_exceeds_limit_for_logging"]

import logging
import warnings

_logger = logging.getLogger(__name__)


_CUSTOM_RUN_ID_LENGTH = 36
_LOGGED_IMAGE_SIZE_LIMIT_MB = 32

BYTES_IN_MB = 1024 * 1024


def custom_run_id_exceeds_length(custom_run_id):
    if custom_run_id and len(custom_run_id) > _CUSTOM_RUN_ID_LENGTH:
        _logger.warning(
            "Given custom_run_id exceeds %s" " characters and it will be ignored.",
            _CUSTOM_RUN_ID_LENGTH,
        )
        return True
    return False


def image_size_exceeds_limit_for_logging(content_size):
    if content_size > _LOGGED_IMAGE_SIZE_LIMIT_MB * BYTES_IN_MB:
        warnings.warn(
            f"You are attempting to log an image that is {content_size / BYTES_IN_MB:.2f}MB large. "
            f"Neptune supports logging images smaller than {_LOGGED_IMAGE_SIZE_LIMIT_MB}MB. "
            "Resize or increase compression of this image.",
            category=UserWarning,
        )
        return True
    return False



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json

from neptune.legacy.internal.websockets.message import Message


class WebsocketMessageProcessor(object):
    def __init__(self):
        pass

    def run(self, raw_message):
        # Atmosphere framework sends heartbeat messages every minute, we have to ignore them
        if raw_message is not None and not self._is_heartbeat(raw_message):
            message = Message.from_json(json.loads(raw_message))
            self._process_message(message)

    def _process_message(self, message):
        raise NotImplementedError()

    @staticmethod
    def _is_heartbeat(message):
        return message.strip() == ""



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re


class GPUCardIndicesProvider(object):
    def __init__(self, cuda_visible_devices, gpu_card_count):
        self.__cuda_visible_devices = cuda_visible_devices
        self.__gpu_card_count = gpu_card_count
        self.__cuda_visible_devices_regex = r"^-?\d+(,-?\d+)*$"

    def get(self):
        if self.__is_cuda_visible_devices_correct():
            return self.__gpu_card_indices_from_cuda_visible_devices()
        else:
            return list(range(self.__gpu_card_count))

    def __is_cuda_visible_devices_correct(self):
        return self.__cuda_visible_devices is not None and re.match(
            self.__cuda_visible_devices_regex, self.__cuda_visible_devices
        )

    def __gpu_card_indices_from_cuda_visible_devices(self):
        correct_indices = []

        # According to CUDA Toolkit specification.
        # https://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#env-vars
        for gpu_index_str in self.__cuda_visible_devices.split(","):
            gpu_index = int(gpu_index_str)
            if 0 <= gpu_index < self.__gpu_card_count:
                correct_indices.append(gpu_index)
            else:
                break

        return list(set(correct_indices))



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["SeriesValue"]

from typing import (
    Generic,
    TypeVar,
)

T = TypeVar("T")


class SeriesValue(Generic[T]):
    def __init__(self, step: float, value: T, timestamp: float):
        self._step = step
        self._value = value
        self._timestamp = timestamp

    @property
    def step(self) -> float:
        return self._step

    @step.setter
    def step(self, step: float):
        self._step = step

    @property
    def value(self) -> T:
        return self._value

    @value.setter
    def value(self, value: T):
        self._value = value

    @property
    def timestamp(self) -> float:
        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp: float):
        self._timestamp = timestamp



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["Model"]

import os
from typing import (
    Iterable,
    Optional,
)

from neptune.attributes.constants import SYSTEM_NAME_ATTRIBUTE_PATH
from neptune.common.exceptions import NeptuneException
from neptune.envs import CONNECTION_MODE
from neptune.exceptions import (
    InactiveModelException,
    NeedExistingModelForReadOnlyMode,
    NeptuneMissingRequiredInitParameter,
    NeptuneModelKeyAlreadyExistsError,
    NeptuneObjectCreationConflict,
)
from neptune.internal.backends.api_model import ApiExperiment
from neptune.internal.backends.nql import (
    NQLAggregator,
    NQLAttributeOperator,
    NQLAttributeType,
    NQLQueryAggregate,
    NQLQueryAttribute,
)
from neptune.internal.backgroud_job_list import BackgroundJobList
from neptune.internal.container_type import ContainerType
from neptune.internal.id_formats import QualifiedName
from neptune.internal.init.parameters import (
    ASYNC_LAG_THRESHOLD,
    ASYNC_NO_PROGRESS_THRESHOLD,
    DEFAULT_FLUSH_PERIOD,
    DEFAULT_NAME,
    OFFLINE_PROJECT_QUALIFIED_NAME,
)
from neptune.internal.state import ContainerState
from neptune.internal.utils import verify_type
from neptune.internal.utils.ping_background_job import PingBackgroundJob
from neptune.metadata_containers import MetadataContainer
from neptune.metadata_containers.abstract import NeptuneObjectCallback
from neptune.metadata_containers.metadata_containers_table import Table
from neptune.types.mode import Mode


class Model(MetadataContainer):
    """Class for registering a model to neptune.ai and retrieving information from it."""

    container_type = ContainerType.MODEL

    def __init__(
        self,
        with_id: Optional[str] = None,
        *,
        name: Optional[str] = None,
        key: Optional[str] = None,
        project: Optional[str] = None,
        api_token: Optional[str] = None,
        mode: Optional[str] = None,
        flush_period: float = DEFAULT_FLUSH_PERIOD,
        proxies: Optional[dict] = None,
        async_lag_callback: Optional[NeptuneObjectCallback] = None,
        async_lag_threshold: float = ASYNC_LAG_THRESHOLD,
        async_no_progress_callback: Optional[NeptuneObjectCallback] = None,
        async_no_progress_threshold: float = ASYNC_NO_PROGRESS_THRESHOLD,
    ):
        """Initializes a Model object from an existing or new model.

        You can use this to create a new model from code or to perform actions on existing models.

        A Model object is suitable for storing model metadata that is common to all versions (you can use ModelVersion
        objects to track version-specific metadata). It does not track background metrics or logs automatically,
        but you can assign metadata to the Model object just like you can for runs.
        To learn more about model registry, see the docs: https://docs.neptune.ai/model_registry/overview/

        You can also use the Model object as a context manager (see examples).

        Args:
             with_id: The Neptune identifier of an existing model to resume, such as "CLS-PRE".
                The identifier is stored in the object's "sys/id" field.
                If omitted or `None` is passed, a new model is created.
            name: A custom name for the model.
            key: Key for the new model. Required when creating a new model version.
                Used together with the project key to form the model identifier.
                Must be uppercase and unique within the project.
            project: Name of a project in the form `workspace-name/project-name`.
                If None, the value of the NEPTUNE_PROJECT environment variable is used.
            api_token: User's API token.
                If None (default), the value of the NEPTUNE_API_TOKEN environment variable is used.
                Note: To keep your API token secure, save it to the NEPTUNE_API_TOKEN environment variable rather than
                placing it in plain text in the source code.
            mode: Connection mode in which the tracking will work.
                If `None` (default), the value of the NEPTUNE_MODE environment variable is used.
                If no value was set for the environment variable, "async" is used by default.
                Possible values are `async`, `sync`, `offline`, `read-only`, and `debug`.
            flush_period: In the asynchronous (default) connection mode, how often disk flushing is triggered
                (in seconds).
            proxies: Argument passed to HTTP calls made via the Requests library, as dictionary of strings.
                For more information about proxies, see the Requests documentation.
            async_lag_callback: Custom callback which is called if the lag between a queued operation and its
                synchronization with the server exceeds the duration defined by `async_lag_threshold`. The callback
                should take a Model object as the argument and can contain any custom code, such as calling `stop()` on
                the object.
                Note: Instead of using this argument, you can use Neptune's default callback by setting the
                `NEPTUNE_ENABLE_DEFAULT_ASYNC_LAG_CALLBACK` environment variable to `TRUE`.
            async_lag_threshold: In seconds, duration between the queueing and synchronization of an operation.
                If a lag callback (default callback enabled via environment variable or custom callback passed to the
                `async_lag_callback` argument) is enabled, the callback is called when this duration is exceeded.
            async_no_progress_callback: Custom callback which is called if there has been no synchronization progress
                whatsoever for the duration defined by `async_no_progress_threshold`. The callback should take a Model
                object as the argument and can contain any custom code, such as calling `stop()` on the object.
                Note: Instead of using this argument, you can use Neptune's default callback by setting the
                `NEPTUNE_ENABLE_DEFAULT_ASYNC_NO_PROGRESS_CALLBACK` environment variable to `TRUE`.
            async_no_progress_threshold: In seconds, for how long there has been no synchronization progress since the
                object was initialized. If a no-progress callback (default callback enabled via environment variable or
                custom callback passed to the `async_no_progress_callback` argument) is enabled, the callback is called
                when this duration is exceeded.

        Returns:
            Model object that is used to manage the model and log metadata to it.

        Examples:

            >>> import neptune

            Creating a new model:

            >>> model = neptune.init_model(key="PRE")
            >>> model["metadata"] = some_metadata

            >>> # Or initialize with the constructor
            ... model = Model(key="PRE")

            >>> # You can provide the project parameter as an environment variable
            ... # or as an argument to the init_model() function:
            ... model = neptune.init_model(key="PRE", project="workspace-name/project-name")

            >>> # When creating a model, you can give it a name:
            ... model = neptune.init_model(key="PRE", name="Pre-trained model")

            Connecting to an existing model:

            >>> # Initialize existing model with identifier "CLS-PRE"
            ... model = neptune.init_model(with_id="CLS-PRE")

            >>> # To prevent modifications when connecting to an existing model, you can connect in read-only mode
            ... model = neptune.init_model(with_id="CLS-PRE", mode="read-only")

            Using the Model object as context manager:

            >>> with Model(key="PRE") as model:
            ...     model["metadata"] = some_metadata

        For details, see the docs:
            Initializing a model:
                https://docs.neptune.ai/api/neptune#init_model
            Model class reference:
                https://docs.neptune.ai/api/model
        """
        verify_type("with_id", with_id, (str, type(None)))
        verify_type("name", name, (str, type(None)))
        verify_type("key", key, (str, type(None)))
        verify_type("project", project, (str, type(None)))
        verify_type("mode", mode, (str, type(None)))

        self._key: Optional[str] = key
        self._with_id: Optional[str] = with_id
        self._name: Optional[str] = DEFAULT_NAME if with_id is None and name is None else name

        # make mode proper Enum instead of string
        mode = Mode(mode or os.getenv(CONNECTION_MODE) or Mode.ASYNC.value)

        if mode == Mode.OFFLINE:
            raise NeptuneException("Model can't be initialized in OFFLINE mode")

        if mode == Mode.DEBUG:
            project = OFFLINE_PROJECT_QUALIFIED_NAME

        super().__init__(
            project=project,
            api_token=api_token,
            mode=mode,
            flush_period=flush_period,
            proxies=proxies,
            async_lag_callback=async_lag_callback,
            async_lag_threshold=async_lag_threshold,
            async_no_progress_callback=async_no_progress_callback,
            async_no_progress_threshold=async_no_progress_threshold,
        )

    def _get_or_create_api_object(self) -> ApiExperiment:
        project_workspace = self._project_api_object.workspace
        project_name = self._project_api_object.name
        project_qualified_name = f"{project_workspace}/{project_name}"

        if self._with_id is not None:
            # with_id (resume existing model) has priority over key (creating a new model)
            #  additional creation parameters (e.g. name) are simply ignored in this scenario
            return self._backend.get_metadata_container(
                container_id=QualifiedName(project_qualified_name + "/" + self._with_id),
                expected_container_type=self.container_type,
            )
        elif self._key is not None:
            if self._mode == Mode.READ_ONLY:
                raise NeedExistingModelForReadOnlyMode()

            try:
                return self._backend.create_model(project_id=self._project_api_object.id, key=self._key)
            except NeptuneObjectCreationConflict as e:
                base_url = self._backend.get_display_address()
                raise NeptuneModelKeyAlreadyExistsError(
                    model_key=self._key,
                    models_tab_url=f"{base_url}/{project_workspace}/{project_name}/models",
                ) from e
        else:
            raise NeptuneMissingRequiredInitParameter(
                parameter_name="key",
                called_function="init_model",
            )

    def _prepare_background_jobs(self) -> BackgroundJobList:
        return BackgroundJobList([PingBackgroundJob()])

    def _write_initial_attributes(self):
        if self._name is not None:
            self[SYSTEM_NAME_ATTRIBUTE_PATH] = self._name

    def _raise_if_stopped(self):
        if self._state == ContainerState.STOPPED:
            raise InactiveModelException(label=self._sys_id)

    def get_url(self) -> str:
        """Returns the URL that can be accessed within the browser"""
        return self._backend.get_model_url(
            model_id=self._id,
            workspace=self._workspace,
            project_name=self._project_name,
            sys_id=self._sys_id,
        )

    def fetch_model_versions_table(self, *, columns: Optional[Iterable[str]] = None) -> Table:
        """Retrieve all versions of the given model.

        Args:
            columns: Names of columns to include in the table, as a list of namespace or field names.
                The Neptune ID ("sys/id") is included automatically.
                Examples:
                    Fields: `["params/lr", "params/batch", "val/acc"]` - these fields are included as columns.
                    Namespaces: `["params", "val"]` - all the fields inside the namespaces are included as columns.
                If `None` (default), all the columns of the model versions table are included.

        Returns:
            `Table` object containing `ModelVersion` objects that match the specified criteria.

            Use `to_pandas()` to convert it to a pandas DataFrame.

        Examples:
            >>> import neptune

            >>> # Initialize model with the ID "CLS-FOREST"
            ... model = neptune.init_model(with_id="CLS-FOREST")

            >>> # Fetch the metadata of all model versions as a pandas DataFrame
            ... model_versions_df = model.fetch_model_versions_table().to_pandas()

            >>> # Fetch the metadata of all model versions as a pandas DataFrame,
            ... # including only the fields "params/lr" and "val/loss" as columns:
            ... model_versions = model.fetch_model_versions_table(columns=["params/lr", "val/loss"])
            ... model_versions_df = model_versions.to_pandas()

            >>> # Sort model versions by size
            ... model_versions_df = model_versions_df.sort_values(by="sys/size")

            >>> # Sort model versions by creation time
            ... model_versions_df = model_versions_df.sort_values(by="sys/creation_time", ascending=False)

            >>> # Extract the last model version ID
            ... last_model_version_id = model_versions_df["sys/id"].values[0]

        See also the API referene:
            https://docs.neptune.ai/api/model/#fetch_model_versions_table
        """
        return MetadataContainer._fetch_entries(
            self,
            child_type=ContainerType.MODEL_VERSION,
            query=NQLQueryAggregate(
                items=[
                    NQLQueryAttribute(
                        name="sys/model_id",
                        value=self._sys_id,
                        operator=NQLAttributeOperator.EQUALS,
                        type=NQLAttributeType.STRING,
                    ),
                    NQLQueryAttribute(
                        name="sys/trashed",
                        type=NQLAttributeType.BOOLEAN,
                        operator=NQLAttributeOperator.EQUALS,
                        value=False,
                    ),
                ],
                aggregator=NQLAggregator.AND,
            ),
            columns=columns,
        )



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["get_batches"]

from itertools import (
    chain,
    islice,
)
from typing import (
    Iterable,
    List,
    TypeVar,
)

T = TypeVar("T")


def get_batches(iterable: Iterable[T], *, batch_size: int) -> Iterable[List[T]]:
    assert batch_size > 0

    source_iter = iter(iterable)
    while True:
        # return consequent slices of `batch_size` elements
        slices = islice(source_iter, batch_size)
        try:
            first_from_slice = next(slices)
        except StopIteration:
            # but if there's nothing to return in last slice, close generator instead of returning empty list
            return
        yield list(chain([first_from_slice], slices))



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "API_TOKEN_ENV_NAME",
    "CONNECTION_MODE",
    "PROJECT_ENV_NAME",
    "CUSTOM_RUN_ID_ENV_NAME",
    "MONITORING_NAMESPACE",
    "NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE",
    "NEPTUNE_NOTEBOOK_ID",
    "NEPTUNE_NOTEBOOK_PATH",
    "NEPTUNE_RETRIES_TIMEOUT_ENV",
    "NEPTUNE_SYNC_BATCH_TIMEOUT_ENV",
    "NEPTUNE_SUBPROCESS_KILL_TIMEOUT",
    "NEPTUNE_FETCH_TABLE_STEP_SIZE",
    "NEPTUNE_SYNC_AFTER_STOP_TIMEOUT",
    "NEPTUNE_REQUEST_TIMEOUT",
    "NEPTUNE_ENABLE_DEFAULT_ASYNC_LAG_CALLBACK",
    "NEPTUNE_ENABLE_DEFAULT_ASYNC_NO_PROGRESS_CALLBACK",
]

from neptune.common.envs import (
    API_TOKEN_ENV_NAME,
    NEPTUNE_RETRIES_TIMEOUT_ENV,
)

CONNECTION_MODE = "NEPTUNE_MODE"

PROJECT_ENV_NAME = "NEPTUNE_PROJECT"

CUSTOM_RUN_ID_ENV_NAME = "NEPTUNE_CUSTOM_RUN_ID"

MONITORING_NAMESPACE = "NEPTUNE_MONITORING_NAMESPACE"

NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE = "NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE"

NEPTUNE_NOTEBOOK_ID = "NEPTUNE_NOTEBOOK_ID"

NEPTUNE_NOTEBOOK_PATH = "NEPTUNE_NOTEBOOK_PATH"

NEPTUNE_SYNC_BATCH_TIMEOUT_ENV = "NEPTUNE_SYNC_BATCH_TIMEOUT"

NEPTUNE_SUBPROCESS_KILL_TIMEOUT = "NEPTUNE_SUBPROCESS_KILL_TIMEOUT"

NEPTUNE_FETCH_TABLE_STEP_SIZE = "NEPTUNE_FETCH_TABLE_STEP_SIZE"

NEPTUNE_SYNC_AFTER_STOP_TIMEOUT = "NEPTUNE_SYNC_AFTER_STOP_TIMEOUT"

NEPTUNE_REQUEST_TIMEOUT = "NEPTUNE_REQUEST_TIMEOUT"

NEPTUNE_ENABLE_DEFAULT_ASYNC_LAG_CALLBACK = "NEPTUNE_ENABLE_DEFAULT_ASYNC_LAG_CALLBACK"

NEPTUNE_ENABLE_DEFAULT_ASYNC_NO_PROGRESS_CALLBACK = "NEPTUNE_ENABLE_DEFAULT_ASYNC_NO_PROGRESS_CALLBACK"

S3_ENDPOINT_URL = "S3_ENDPOINT_URL"



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "ANONYMOUS_API_TOKEN",
    "NEPTUNE_DATA_DIRECTORY",
    "OFFLINE_DIRECTORY",
    "ASYNC_DIRECTORY",
    "SYNC_DIRECTORY",
    "OFFLINE_NAME_PREFIX",
    "MAX_32_BIT_INT",
    "MIN_32_BIT_INT",
]

"""Constants used by Neptune"""


ANONYMOUS_API_TOKEN = "ANONYMOUS"

NEPTUNE_DATA_DIRECTORY = ".neptune"

OFFLINE_DIRECTORY = "offline"
ASYNC_DIRECTORY = "async"
SYNC_DIRECTORY = "sync"

OFFLINE_NAME_PREFIX = "offline/"

MAX_32_BIT_INT = 2147483647
MIN_32_BIT_INT = -2147483648



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["parse_path", "path_to_str", "join_paths"]

from typing import List


def _remove_empty_paths(paths: List[str]) -> List[str]:
    return list(filter(bool, paths))


def parse_path(path: str) -> List[str]:
    return _remove_empty_paths(path.split("/"))


def path_to_str(path: List[str]) -> str:
    return "/".join(_remove_empty_paths(path))


def join_paths(*paths: str) -> str:
    return "/".join(_remove_empty_paths(list(paths)))



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "MetadataInconsistency",
    "AttributeType",
    "AttributeWithProperties",
    "NeptuneBackend",
    "ContainerType",
    "LeaderboardEntry",
    "LeaderboardHandler",
    "RunsTable",
    "RunsTableEntry",
]

# backwards compatibility
from neptune.exceptions import MetadataInconsistency
from neptune.internal.backends.api_model import (
    AttributeType,
    AttributeWithProperties,
)
from neptune.internal.backends.neptune_backend import NeptuneBackend
from neptune.internal.container_type import ContainerType
from neptune.metadata_containers.metadata_containers_table import (
    LeaderboardEntry,
    LeaderboardHandler,
)
from neptune.metadata_containers.metadata_containers_table import Table as RunsTable
from neptune.metadata_containers.metadata_containers_table import TableEntry as RunsTableEntry



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from functools import wraps
from typing import Optional

from neptune.common.warnings import warn_once
from neptune.exceptions import NeptuneParametersCollision

__all__ = ["deprecated", "deprecated_parameter"]


def deprecated(*, alternative: Optional[str] = None):
    def deco(func):
        @wraps(func)
        def inner(*args, **kwargs):
            additional_info = f", use `{alternative}` instead" if alternative else " and will be removed"

            warn_once(
                message=f"`{func.__name__}` is deprecated{additional_info}."
                f" We'll end support of it in next major release."
            )

            return func(*args, **kwargs)

        return inner

    return deco


def deprecated_parameter(*, deprecated_kwarg_name, required_kwarg_name):
    def deco(f):
        @wraps(f)
        def inner(*args, **kwargs):
            if deprecated_kwarg_name in kwargs:
                if required_kwarg_name in kwargs:
                    raise NeptuneParametersCollision(required_kwarg_name, deprecated_kwarg_name, method_name=f.__name__)

                warn_once(
                    message=f"Parameter `{deprecated_kwarg_name}` is deprecated, use `{required_kwarg_name}` instead."
                    " We'll end support of it in next major release."
                )

                kwargs[required_kwarg_name] = kwargs[deprecated_kwarg_name]
                del kwargs[deprecated_kwarg_name]

            return f(*args, **kwargs)

        return inner

    return deco



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["QueueElement", "DiskQueue"]

import json
import logging
import os
import shutil
import threading
from dataclasses import dataclass
from glob import glob
from pathlib import Path
from typing import (
    Callable,
    Generic,
    List,
    Optional,
    Tuple,
    TypeVar,
)

from neptune.exceptions import MalformedOperation
from neptune.internal.utils.json_file_splitter import JsonFileSplitter
from neptune.internal.utils.sync_offset_file import SyncOffsetFile

T = TypeVar("T")

_logger = logging.getLogger(__name__)


@dataclass
class QueueElement(Generic[T]):
    obj: T
    ver: int
    size: int


class DiskQueue(Generic[T]):
    # NOTICE: This class is thread-safe as long as there is only one consumer and one producer.
    DEFAULT_MAX_BATCH_SIZE_BYTES = 100 * 1024**2

    def __init__(
        self,
        dir_path: Path,
        to_dict: Callable[[T], dict],
        from_dict: Callable[[dict], T],
        lock: threading.RLock,
        max_file_size: int = 64 * 1024**2,
        max_batch_size_bytes: int = None,
    ):
        self._dir_path = dir_path.resolve()
        self._to_dict = to_dict
        self._from_dict = from_dict
        self._max_file_size = max_file_size
        self._max_batch_size_bytes = max_batch_size_bytes or int(
            os.environ.get("NEPTUNE_MAX_BATCH_SIZE_BYTES") or str(self.DEFAULT_MAX_BATCH_SIZE_BYTES)
        )

        try:
            os.makedirs(self._dir_path)
        except FileExistsError:
            pass

        self._last_ack_file = SyncOffsetFile(dir_path / "last_ack_version", default=0)
        self._last_put_file = SyncOffsetFile(dir_path / "last_put_version", default=0)

        (
            self._read_file_version,
            self._write_file_version,
        ) = self._get_first_and_last_log_file_version()
        self._writer = open(self._get_log_file(self._write_file_version), "a")
        self._reader = JsonFileSplitter(self._get_log_file(self._read_file_version))
        self._file_size = 0
        self._should_skip_to_ack = True

        self._empty_cond = threading.Condition(lock)

    def put(self, obj: T) -> int:
        version = self._last_put_file.read_local() + 1
        _json = json.dumps(self._serialize(obj, version))
        if self._file_size + len(_json) > self._max_file_size:
            old_writer = self._writer
            self._writer = open(self._get_log_file(version), "a")
            old_writer.flush()
            old_writer.close()
            self._file_size = 0
            self._write_file_version = version
        self._writer.write(_json + "\n")
        self._last_put_file.write(version)
        self._file_size += len(_json) + 1
        return version

    def get(self) -> Optional[QueueElement[T]]:
        if self._should_skip_to_ack:
            return self._skip_and_get()
        else:
            return self._get()

    def _skip_and_get(self) -> Optional[QueueElement[T]]:
        ack_version = self._last_ack_file.read_local()
        while True:
            top_element = self._get()
            if top_element is None:
                return None
            if top_element.ver > ack_version:
                self._should_skip_to_ack = False
                if top_element.ver > ack_version + 1:
                    _logger.warning(
                        "Possible data loss. Last acknowledged operation version: %d, next: %d",
                        ack_version,
                        top_element.ver,
                    )
                return top_element

    def _get(self) -> Optional[QueueElement[T]]:
        _json, size = self._reader.get_with_size()
        if not _json:
            if self._read_file_version >= self._write_file_version:
                return None
            self._reader.close()
            self._read_file_version = self._next_log_file_version(self._read_file_version)
            self._reader = JsonFileSplitter(self._get_log_file(self._read_file_version))
            # It is safe. Max recursion level is 2.
            return self._get()
        try:
            obj, ver = self._deserialize(_json)
            return QueueElement[T](obj, ver, size)
        except Exception as e:
            raise MalformedOperation from e

    def get_batch(self, size: int) -> List[QueueElement[T]]:
        if self._should_skip_to_ack:
            first = self._skip_and_get()
        else:
            first = self._get()
        if not first:
            return []

        ret = [first]
        cur_batch_size = first.size
        for _ in range(0, size - 1):
            if cur_batch_size >= self._max_batch_size_bytes:
                break
            next_obj = self._get()
            if not next_obj:
                break

            cur_batch_size += next_obj.size
            ret.append(next_obj)
        return ret

    def flush(self):
        self._writer.flush()
        self._last_ack_file.flush()
        self._last_put_file.flush()

    def close(self):
        self._reader.close()
        self._writer.close()
        self._last_ack_file.close()
        self._last_put_file.close()

    def cleanup_if_empty(self) -> None:
        """
        Remove underlying files if queue is empty
        """
        if self.is_empty():
            self._remove_data()

    def _remove_data(self):
        path = self._dir_path
        shutil.rmtree(path, ignore_errors=True)

        parent = path.parent

        try:
            files = os.listdir(parent)
        except FileNotFoundError:
            files = []

        if len(files) == 0:
            try:
                os.rmdir(parent)
            except OSError:
                _logger.info(f"Cannot remove directory: {parent}")

    def wait_for_empty(self, seconds: Optional[float] = None) -> bool:
        with self._empty_cond:
            return self._empty_cond.wait_for(self.is_empty, timeout=seconds)

    def ack(self, version: int) -> None:
        self._last_ack_file.write(version)

        log_versions = self._get_all_log_file_versions()
        for i in range(0, len(log_versions) - 1):
            if log_versions[i + 1] <= version:
                filename = self._get_log_file(log_versions[i])
                try:
                    os.remove(filename)
                except FileNotFoundError:
                    # not really a problem
                    pass
                except Exception:
                    _logger.exception("Cannot remove queue file %s", filename)
            else:
                break

        with self._empty_cond:
            if self.is_empty():
                self._empty_cond.notify_all()

    def is_empty(self) -> bool:
        return self.size() == 0

    def size(self) -> int:
        return self._last_put_file.read_local() - self._last_ack_file.read_local()

    def _get_log_file(self, index: int) -> str:
        return "{}/data-{}.log".format(self._dir_path, index)

    def _get_all_log_file_versions(self):
        log_files = glob("{}/data-*.log".format(self._dir_path))
        if not log_files:
            return 1, 1
        return sorted([int(file[len(str(self._dir_path)) + 6 : -4]) for file in log_files])

    def _get_first_and_last_log_file_version(self) -> (int, int):
        log_versions = self._get_all_log_file_versions()
        return min(log_versions), max(log_versions)

    def _next_log_file_version(self, version: int) -> int:
        log_versions = self._get_all_log_file_versions()
        for i, val in enumerate(log_versions):
            if val == version:
                return log_versions[i + 1]
        raise ValueError("Missing log file with version > {}".format(version))

    def _serialize(self, obj: T, version: int) -> dict:
        return {"obj": self._to_dict(obj), "version": version}

    def _deserialize(self, data: dict) -> Tuple[T, int]:
        return self._from_dict(data["obj"]), data["version"]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.flush()
        self.close()
        self.cleanup_if_empty()



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["Daemon"]

import abc
import functools
import threading
import time
from enum import Enum

from neptune.common.exceptions import NeptuneConnectionLostException
from neptune.internal.utils.logger import logger


class Daemon(threading.Thread):
    class DaemonState(Enum):
        INIT = 1
        WORKING = 2
        PAUSING = 3
        PAUSED = 4
        INTERRUPTED = 5
        STOPPED = 6

    def __init__(self, sleep_time: float, name):
        super().__init__(daemon=True, name=name)
        self._sleep_time = sleep_time
        self._state: Daemon.DaemonState = Daemon.DaemonState.INIT
        self._wait_condition = threading.Condition()
        self.last_backoff_time = 0  # used only with ConnectionRetryWrapper decorator

    def interrupt(self):
        with self._wait_condition:
            self._state = Daemon.DaemonState.INTERRUPTED
            self._wait_condition.notify_all()

    def pause(self):
        with self._wait_condition:
            if self._state != Daemon.DaemonState.PAUSED:
                if not self._is_interrupted():
                    self._state = Daemon.DaemonState.PAUSING
                self._wait_condition.notify_all()
                self._wait_condition.wait_for(lambda: self._state != Daemon.DaemonState.PAUSING)

    def resume(self):
        with self._wait_condition:
            if not self._is_interrupted():
                self._state = Daemon.DaemonState.WORKING
            self._wait_condition.notify_all()

    def wake_up(self):
        with self._wait_condition:
            self._wait_condition.notify_all()

    def disable_sleep(self):
        self._sleep_time = 0

    def is_running(self) -> bool:
        with self._wait_condition:
            return self._state in (
                Daemon.DaemonState.WORKING,
                Daemon.DaemonState.PAUSING,
                Daemon.DaemonState.PAUSED,
            )

    def _is_interrupted(self) -> bool:
        with self._wait_condition:
            return self._state in (Daemon.DaemonState.INTERRUPTED, Daemon.DaemonState.STOPPED)

    def run(self):
        with self._wait_condition:
            if not self._is_interrupted():
                self._state = Daemon.DaemonState.WORKING
        try:
            while not self._is_interrupted():
                with self._wait_condition:
                    if self._state == Daemon.DaemonState.PAUSING:
                        self._state = Daemon.DaemonState.PAUSED
                        self._wait_condition.notify_all()
                        self._wait_condition.wait_for(lambda: self._state != Daemon.DaemonState.PAUSED)

                if self._state == Daemon.DaemonState.WORKING:
                    self.work()
                    with self._wait_condition:
                        if self._sleep_time > 0 and self._state == Daemon.DaemonState.WORKING:
                            self._wait_condition.wait(timeout=self._sleep_time)
        finally:
            with self._wait_condition:
                self._state = Daemon.DaemonState.STOPPED
                self._wait_condition.notify_all()

    @abc.abstractmethod
    def work(self):
        pass

    class ConnectionRetryWrapper:
        INITIAL_RETRY_BACKOFF = 2
        MAX_RETRY_BACKOFF = 120

        def __init__(self, kill_message):
            self.kill_message = kill_message

        def __call__(self, func):
            @functools.wraps(func)
            def wrapper(self_: Daemon, *args, **kwargs):
                while not self_._is_interrupted():
                    try:
                        result = func(self_, *args, **kwargs)
                        if self_.last_backoff_time > 0:
                            self_.last_backoff_time = 0
                            logger.info("Communication with Neptune restored!")
                        return result
                    except NeptuneConnectionLostException as e:
                        if self_.last_backoff_time == 0:
                            logger.warning(
                                "Experiencing connection interruptions."
                                " Will try to reestablish communication with Neptune."
                                " Internal exception was: %s",
                                e.cause.__class__.__name__,
                            )
                            self_.last_backoff_time = self.INITIAL_RETRY_BACKOFF
                        else:
                            self_.last_backoff_time = min(self_.last_backoff_time * 2, self.MAX_RETRY_BACKOFF)
                        time.sleep(self_.last_backoff_time)
                    except Exception:
                        logger.error(
                            "Unexpected error occurred in Neptune background thread: %s",
                            self.kill_message,
                        )
                        raise

            return wrapper



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import warnings

import click
import pkg_resources

from neptune.cli.commands import (
    clear,
    status,
    sync,
)


@click.group()
def main():
    pass


main.add_command(sync)
main.add_command(status)
main.add_command(clear)

plugins = {entry_point.name: entry_point for entry_point in pkg_resources.iter_entry_points("neptune.plugins")}

for name, entry_point in plugins.items():
    # loading an entry_point may fail and this
    # will cause all CLI commands to fail.
    # So, we load the plug-ins in try and except block.
    try:
        loaded_plugin = entry_point.load()
    except Exception as e:
        warnings.warn(f"Failed to load neptune plug-in `{name}` with exception: {e}")
    main.add_command(loaded_plugin, name)



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import socket
import sys

import click
from bravado.client import SwaggerClient
from bravado_core.formatter import SwaggerFormat
from packaging import version
from six.moves import urllib

from neptune.legacy.exceptions import (
    CannotResolveHostname,
    DeprecatedApiToken,
    UnsupportedClientVersion,
)
from neptune.legacy.internal.api_clients.client_config import (
    ClientConfig,
    MultipartConfig,
)
from neptune.legacy.internal.api_clients.hosted_api_clients.utils import legacy_with_api_exceptions_handler

_logger = logging.getLogger(__name__)

uuid_format = SwaggerFormat(
    format="uuid",
    to_python=lambda x: x,
    to_wire=lambda x: x,
    validate=lambda x: None,
    description="",
)


class HostedNeptuneMixin:
    """Mixin containing operation common for both backend and leaderboard api clients"""

    @legacy_with_api_exceptions_handler
    def _get_swagger_client(self, url, http_client):
        return SwaggerClient.from_url(
            url,
            config=dict(
                validate_swagger_spec=False,
                validate_requests=False,
                validate_responses=False,
                formats=[uuid_format],
            ),
            http_client=http_client,
        )

    @staticmethod
    def _get_client_config_args(api_token):
        return dict(X_Neptune_Api_Token=api_token)

    @legacy_with_api_exceptions_handler
    def _create_client_config(self, api_token, backend_client):
        client_config_args = self._get_client_config_args(api_token)
        config = backend_client.api.getClientConfig(**client_config_args).response().result

        if hasattr(config, "pyLibVersions"):
            min_recommended = getattr(config.pyLibVersions, "minRecommendedVersion", None)
            min_compatible = getattr(config.pyLibVersions, "minCompatibleVersion", None)
            max_compatible = getattr(config.pyLibVersions, "maxCompatibleVersion", None)
        else:
            click.echo(
                "ERROR: This client version is not supported by your Neptune instance. Please contant Neptune support.",
                sys.stderr,
            )
            raise UnsupportedClientVersion(self.client_lib_version, None, "0.4.111")

        multipart_upload_config_obj = getattr(config, "multiPartUpload", None)
        has_multipart_upload = getattr(multipart_upload_config_obj, "enabled", False)
        if not has_multipart_upload:
            multipart_upload_config = None
        else:
            min_chunk_size = getattr(multipart_upload_config_obj, "minChunkSize")
            max_chunk_size = getattr(multipart_upload_config_obj, "maxChunkSize")
            max_chunk_count = getattr(multipart_upload_config_obj, "maxChunkCount")
            max_single_part_size = getattr(multipart_upload_config_obj, "maxSinglePartSize")
            multipart_upload_config = MultipartConfig(
                min_chunk_size, max_chunk_size, max_chunk_count, max_single_part_size
            )

        return ClientConfig(
            api_url=config.apiUrl,
            display_url=config.applicationUrl,
            min_recommended_version=version.parse(min_recommended) if min_recommended else None,
            min_compatible_version=version.parse(min_compatible) if min_compatible else None,
            max_compatible_version=version.parse(max_compatible) if max_compatible else None,
            multipart_config=multipart_upload_config,
        )

    def _verify_host_resolution(self, api_url, app_url):
        host = urllib.parse.urlparse(api_url).netloc.split(":")[0]
        try:
            socket.gethostbyname(host)
        except socket.gaierror:
            if self.credentials.api_url_opt is None:
                raise DeprecatedApiToken(urllib.parse.urlparse(app_url).netloc)
            else:
                raise CannotResolveHostname(host)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import time
from collections import namedtuple
from enum import Enum
from typing import List

from neptune.legacy.exceptions import NeptuneException

ChannelNameWithTypeAndNamespace = namedtuple(
    "ChannelNameWithType",
    ["channel_id", "channel_name", "channel_type", "channel_namespace"],
)


class ChannelType(Enum):
    TEXT = "text"
    NUMERIC = "numeric"
    IMAGE = "image"


class ChannelValueType(Enum):
    TEXT_VALUE = "text_value"
    NUMERIC_VALUE = "numeric_value"
    IMAGE_VALUE = "image_value"


class ChannelNamespace(Enum):
    USER = "user"
    SYSTEM = "system"


class ChannelValue(object):
    def __init__(self, x, y, ts):
        self._x = x
        self._y = y
        if ts is None:
            ts = time.time()
        self._ts = ts

    @property
    def ts(self):
        return self._ts

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def value(self):
        return self.y.get(self.value_type.value)

    @property
    def value_type(self) -> ChannelValueType:
        """We expect that exactly one of `y` values is not None, and according to that we try to determine value type"""
        unique_channel_value_types = set(
            [ch_value_type for ch_value_type in ChannelValueType if self.y.get(ch_value_type.value) is not None]
        )
        if len(unique_channel_value_types) > 1:
            raise NeptuneException(f"There are mixed value types in {self}")
        if not unique_channel_value_types:
            raise NeptuneException(f"Can't determine type of {self}")

        return next(iter(unique_channel_value_types))

    def __str__(self):
        return "ChannelValue(x={},y={},ts={})".format(self.x, self.y, self.ts)

    def __repr__(self):
        return str(self)

    def __eq__(self, o):
        return self.__dict__ == o.__dict__


class ChannelIdWithValues:
    def __init__(self, channel_id, channel_name, channel_type, channel_namespace, channel_values):
        self._channel_id = channel_id
        self._channel_name = channel_name
        self._channel_type = channel_type
        self._channel_namespace = channel_namespace
        self._channel_values = channel_values

    @property
    def channel_id(self) -> str:
        return self._channel_id

    @property
    def channel_name(self) -> str:
        return self._channel_name

    @property
    def channel_values(self) -> List[ChannelValue]:
        return self._channel_values

    @property
    def channel_type(self) -> ChannelValueType:
        if self._channel_type == ChannelType.NUMERIC.value:
            return ChannelValueType.NUMERIC_VALUE
        elif self._channel_type == ChannelType.TEXT.value:
            return ChannelValueType.TEXT_VALUE
        elif self._channel_type == ChannelType.IMAGE.value:
            return ChannelValueType.IMAGE_VALUE
        else:
            raise NeptuneException(f"Unknown channel type: {self._channel_type}")

    @property
    def channel_namespace(self) -> ChannelNamespace:
        return self._channel_namespace

    def __eq__(self, other):
        return self.channel_id == other.channel_id and self.channel_values == other.channel_values

    def __gt__(self, other):
        return hash(self.channel_id) < hash(other.channel_id)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "FileSet",
]

from typing import (
    TYPE_CHECKING,
    Iterable,
    List,
    TypeVar,
    Union,
)

from neptune.internal.utils import (
    verify_collection_type,
    verify_type,
)
from neptune.types.value import Value

if TYPE_CHECKING:
    from neptune.types.value_visitor import ValueVisitor

Ret = TypeVar("Ret")


class FileSet(Value):
    def __init__(self, file_globs: Union[str, Iterable[str]]):
        verify_type("file_globs", file_globs, (str, Iterable))
        if isinstance(file_globs, str):
            file_globs = [file_globs]
        else:
            verify_collection_type("file_globs", file_globs, str)
        self.file_globs: List[str] = list(file_globs)

    def accept(self, visitor: "ValueVisitor[Ret]") -> Ret:
        return visitor.visit_file_set(self)

    def __str__(self):
        return "FileSet({})".format(str(self.file_globs))



__all__ = ["ANONYMOUS_API_TOKEN_CONTENT", "DO_NOT_TRACK_GIT_REPOSITORY"]

ANONYMOUS_API_TOKEN_CONTENT = (
    "eyJhcGlfYWRkcmVzcyI6Imh0dHBzOi8vYXBwLm5lcHR1bmUuYWkiLCJhcGlfdXJsIjoiaHR0cHM6Ly9hcHAubmVwdHVuZS"
    "5haSIsImFwaV9rZXkiOiJiNzA2YmM4Zi03NmY5LTRjMmUtOTM5ZC00YmEwMzZmOTMyZTQifQo="
)
DO_NOT_TRACK_GIT_REPOSITORY = "DO_NOT_TRACK_GIT_REPOSITORY"



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = [
    "get_project_list",
    "create_project",
    "delete_project",
    "get_project_member_list",
    "add_project_member",
    "remove_project_member",
    "get_workspace_member_list",
    "invite_to_workspace",
    "WorkspaceMemberRole",
    "add_project_service_account",
    "remove_project_service_account",
    "get_project_service_account_list",
    "get_workspace_service_account_list",
    "trash_objects",
]

import os
from typing import (
    Dict,
    Iterable,
    List,
    Optional,
    Union,
)

from bravado.exception import (
    HTTPBadRequest,
    HTTPConflict,
    HTTPForbidden,
    HTTPNotFound,
    HTTPUnprocessableEntity,
)

from neptune.common.backends.utils import with_api_exceptions_handler
from neptune.common.envs import API_TOKEN_ENV_NAME
from neptune.internal.backends.hosted_client import (
    DEFAULT_REQUEST_KWARGS,
    create_backend_client,
    create_http_client_with_auth,
    create_leaderboard_client,
)
from neptune.internal.backends.swagger_client_wrapper import SwaggerClientWrapper
from neptune.internal.backends.utils import (
    parse_validation_errors,
    ssl_verify,
)
from neptune.internal.credentials import Credentials
from neptune.internal.id_formats import QualifiedName
from neptune.internal.utils import (
    verify_collection_type,
    verify_type,
)
from neptune.internal.utils.iteration import get_batches
from neptune.internal.utils.logger import logger
from neptune.management.exceptions import (
    AccessRevokedOnDeletion,
    AccessRevokedOnMemberRemoval,
    AccessRevokedOnServiceAccountRemoval,
    BadRequestException,
    ProjectAlreadyExists,
    ProjectNotFound,
    ServiceAccountAlreadyHasAccess,
    ServiceAccountNotExistsOrWithoutAccess,
    ServiceAccountNotFound,
    UserAlreadyHasAccess,
    UserAlreadyInvited,
    UserNotExistsOrWithoutAccess,
    WorkspaceNotFound,
    WorkspaceOrUserNotFound,
)
from neptune.management.internal.dto import (
    ProjectMemberRoleDTO,
    ProjectVisibilityDTO,
    ServiceAccountDTO,
    WorkspaceMemberRoleDTO,
)
from neptune.management.internal.types import ProjectVisibility
from neptune.management.internal.utils import (
    WorkspaceMemberRole,
    extract_project_and_workspace,
    normalize_project_name,
)

TRASH_BATCH_SIZE = 100


def _get_token(api_token: Optional[str] = None) -> str:
    return api_token or os.getenv(API_TOKEN_ENV_NAME)


def _get_http_client_and_config(api_token: Optional[str] = None):
    credentials = Credentials.from_token(api_token=_get_token(api_token=api_token))
    http_client, client_config = create_http_client_with_auth(
        credentials=credentials, ssl_verify=ssl_verify(), proxies={}
    )
    return http_client, client_config


def _get_backend_client(api_token: Optional[str] = None) -> SwaggerClientWrapper:
    http_client, client_config = _get_http_client_and_config(api_token)
    return create_backend_client(client_config=client_config, http_client=http_client)


def _get_leaderboard_client(api_token: Optional[str] = None) -> SwaggerClientWrapper:
    http_client, client_config = _get_http_client_and_config(api_token)
    return create_leaderboard_client(client_config=client_config, http_client=http_client)


def get_project_list(*, api_token: Optional[str] = None) -> List[str]:
    """Lists projects that the account has access to.

    Args:
        api_token: Account's API token.
            If None, the value of the NEPTUNE_API_TOKEN environment variable is used.
            Note: To keep your token secure, use the NEPTUNE_API_TOKEN environment variable rather than placing your
            API token in plain text in your source code.

    Returns:
        List of project names in the form 'workspace-name/project-name'.

    Example:
        >>> from neptune import management
        >>> management.get_project_list()

    You may also want to check the management API reference:
    https://docs.neptune.ai/api/management
    """
    verify_type("api_token", api_token, (str, type(None)))
    backend_client = _get_backend_client(api_token=api_token)
    params = {
        "userRelation": "viewerOrHigher",
        "sortBy": ["lastViewed"],
        **DEFAULT_REQUEST_KWARGS,
    }
    projects = _get_projects(backend_client, params)

    return [normalize_project_name(name=project.name, workspace=project.organizationName) for project in projects]


@with_api_exceptions_handler
def _get_projects(backend_client, params) -> List:
    return backend_client.api.listProjects(**params).response().result.entries


def create_project(
    name: str,
    *,
    key: Optional[str] = None,
    workspace: Optional[str] = None,
    visibility: str = ProjectVisibility.PRIVATE,
    description: Optional[str] = None,
    api_token: Optional[str] = None,
) -> str:
    """Creates a new project in a Neptune workspace.

    Args:
        name: The name for the project in Neptune. Can contain letters and hyphens. For example, 'classification'.
            If you leave out the workspace argument, include the workspace name here,
            in the form 'workspace-name/project-name'. For example, 'ml-team/classification'.
        key: Project identifier. Must contain 1-10 upper case letters or numbers (at least one letter).
            For example, 'CLS2'. If you leave it out, Neptune generates a project key for you.
        workspace: Name of your Neptune workspace.
            If None, it will be parsed from the name argument.
        visibility: Level of visibility you want your project to have.
            Can be set to:
             - 'pub': Public project
             - 'priv': Private project
             - 'workspace' (team workspaces only): Accessible to all workspace members
            If None, it will be set to 'priv'.
        description: Project description.
            If None, it will be left empty.
        api_token: Account's API token.
            If None, the value of the NEPTUNE_API_TOKEN environment variable is used.
            Note: To keep your token secure, use the NEPTUNE_API_TOKEN environment variable rather than placing your
            API token in plain text in your source code.

    Returns:
        The name of the new project you created.

    Example:
        >>> from neptune import management
        >>> management.create_project(
        ...     workspace="ml-team",
        ...     name="classification",
        ...     key="CLS",
        ...     visibility="pub",
        ... )
        'ml-team/classification'

    You may also want to check the management API reference:
    https://docs.neptune.ai/api/management
    """
    verify_type("name", name, str)
    verify_type("key", key, (str, type(None)))
    verify_type("workspace", workspace, (str, type(None)))
    verify_type("visibility", visibility, str)
    verify_type("description", description, (str, type(None)))
    verify_type("api_token", api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)
    workspace, name = extract_project_and_workspace(name=name, workspace=workspace)
    project_qualified_name = f"{workspace}/{name}"
    workspace_id = _get_workspace_id(backend_client, workspace)

    params = {
        "projectToCreate": {
            "name": name,
            "description": description,
            "projectKey": key,
            "organizationId": workspace_id,
            "visibility": ProjectVisibilityDTO.from_str(visibility).value,
        },
        **DEFAULT_REQUEST_KWARGS,
    }

    project_response = _create_project(backend_client, project_qualified_name, params)

    return normalize_project_name(
        name=project_response.result.name,
        workspace=project_response.result.organizationName,
    )


def _get_workspace_id(backend_client, workspace) -> str:
    workspaces = _get_workspaces(backend_client)
    workspace_name_to_id = {f"{f.name}": f.id for f in workspaces}
    if workspace not in workspace_name_to_id:
        raise WorkspaceNotFound(workspace=workspace)
    return workspace_name_to_id[workspace]


@with_api_exceptions_handler
def _get_workspaces(backend_client):
    return backend_client.api.listOrganizations(**DEFAULT_REQUEST_KWARGS).response().result


@with_api_exceptions_handler
def _create_project(backend_client, project_qualified_name: str, params: dict):
    try:
        return backend_client.api.createProject(**params).response()
    except HTTPBadRequest as e:
        validation_errors = parse_validation_errors(error=e)
        if "ERR_NOT_UNIQUE" in validation_errors:
            raise ProjectAlreadyExists(name=project_qualified_name) from e
        raise BadRequestException(validation_errors=validation_errors)


@with_api_exceptions_handler
def delete_project(project: str, *, workspace: Optional[str] = None, api_token: Optional[str] = None):
    """Deletes a project from a Neptune workspace.

    To delete projects, the user must be a workspace admin.

    Args:
        project: The name of the project in Neptune in the form 'workspace-name/project-name'.
            If you pass the workspace argument, the name argument should only contain 'project-name'
            instead of 'workspace-name/project-name'.
        workspace: Name of your Neptune workspace. If you specify it,
            change the format of the name argument to 'project-name' instead of 'workspace-name/project-name'.
            If None, it will be parsed from the name argument.
        api_token: User's API token.
            If None, the value of the NEPTUNE_API_TOKEN environment variable is used.
            Note: To keep your token secure, use the NEPTUNE_API_TOKEN environment variable rather than placing your
            API token in plain text in your source code.

    Example:
        >>> from neptune import management
        >>> management.delete_project(project="ml-team/classification")

    You may also want to check the management API reference:
    https://docs.neptune.ai/api/management
    """
    verify_type("project", project, str)
    verify_type("workspace", workspace, (str, type(None)))
    verify_type("api_token", api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)
    project_identifier = normalize_project_name(name=project, workspace=workspace)

    params = {"projectIdentifier": project_identifier, **DEFAULT_REQUEST_KWARGS}

    try:
        backend_client.api.deleteProject(**params).response()
    except HTTPNotFound as e:
        raise ProjectNotFound(name=project_identifier) from e
    except HTTPForbidden as e:
        raise AccessRevokedOnDeletion(name=project_identifier) from e


@with_api_exceptions_handler
def add_project_member(
    project: str,
    username: str,
    role: str,
    *,
    workspace: Optional[str] = None,
    api_token: Optional[str] = None,
):
    """Adds a member to a Neptune project.

    Only project owners can add members.

    Args:
        project: The name of the project in Neptune in the form 'workspace-name/project-name'.
            If you pass the workspace argument, the name argument should only contain 'project-name'
            instead of 'workspace-name/project-name'.
        username: Name of the user to add to the project.
        role: level of permissions the user should have in a project.
            Can be set to:
             - 'viewer': can only view project content and members
             - 'contributor': can view and edit project content and only view members
             - 'owner': can view and edit project content and members
            For more information, see https://docs.neptune.ai/management/roles/
        workspace: Name of your Neptune workspace. If you specify it,
            change the format of the name argument to 'project-name' instead of 'workspace-name/project-name'.
            If None, it will be parsed from the name argument.
        api_token: Account's API token.
            If None, the value of the NEPTUNE_API_TOKEN environment variable is used.
            Note: To keep your token secure, use the NEPTUNE_API_TOKEN environment variable rather than placing your
            API token in plain text in your source code.

    Examples:
        >>> from neptune import management
        >>> management.add_project_member(
        ...     workspace="ml-team",
        ...     project="classification",
        ...     username="johnny",
        ...     role="contributor",
        ... )

    You may also want to check the management API reference:
    https://docs.neptune.ai/api/management
    """
    verify_type("project", project, str)
    verify_type("username", username, str)
    verify_type("role", role, str)
    verify_type("workspace", workspace, (str, type(None)))
    verify_type("api_token", api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)
    project_identifier = normalize_project_name(name=project, workspace=workspace)

    params = {
        "projectIdentifier": project_identifier,
        "member": {
            "userId": username,
            "role": ProjectMemberRoleDTO.from_str(role).value,
        },
        **DEFAULT_REQUEST_KWARGS,
    }

    try:
        backend_client.api.addProjectMember(**params).response()
    except HTTPNotFound as e:
        raise ProjectNotFound(name=project_identifier) from e
    except HTTPConflict as e:
        members = get_project_member_list(project=project, workspace=workspace, api_token=api_token)
        user_role = members.get(username)
        raise UserAlreadyHasAccess(user=username, project=project_identifier, role=user_role) from e


@with_api_exceptions_handler
def get_project_member_list(
    project: str, *, workspace: Optional[str] = None, api_token: Optional[str] = None
) -> Dict[str, str]:
    """Lists members of a Neptune project.

    Args:
        project: The name of the project in Neptune in the form 'workspace-name/project-name'.
            If you pass the workspace argument, the name argument should only contain 'project-name'
            instead of 'workspace-name/project-name'.
        workspace: Name of your Neptune workspace. If you specify it,
            change the format of the name argument to 'project-name' instead of 'workspace-name/project-name'.
            If None, it will be parsed from the name argument.
        api_token: Account's API token.
            If None, the value of the NEPTUNE_API_TOKEN environment variable is used.
            Note: To keep your token secure, use the NEPTUNE_API_TOKEN environment variable rather than placing your
            API token in plain text in your source code.

    Returns:
        Dictionary with usernames as keys and project member roles
        ('owner', 'contributor', 'viewer') as values.

    Example:
        >>> from neptune import management
        >>> management.get_project_member_list(project="ml-team/classification")

    You may also want to check the management API reference:
    https://docs.neptune.ai/api/management
    """
    verify_type("project", project, str)
    verify_type("workspace", workspace, (str, type(None)))
    verify_type("api_token", api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)
    project_identifier = normalize_project_name(name=project, workspace=workspace)

    params = {"projectIdentifier": project_identifier, **DEFAULT_REQUEST_KWARGS}

    try:
        result = backend_client.api.listProjectMembers(**params).response().result
        return {f"{m.registeredMemberInfo.username}": ProjectMemberRoleDTO.to_domain(m.role) for m in result}
    except HTTPNotFound as e:
        raise ProjectNotFound(name=project_identifier) from e


@with_api_exceptions_handler
def remove_project_member(
    project: str,
    username: str,
    *,
    workspace: Optional[str] = None,
    api_token: Optional[str] = None,
):
    """Removes member from a Neptune project.

    Only project owners can remove members.

    Args:
        project: The name of the project in Neptune in the form 'workspace-name/project-name'.
            If you pass the workspace argument, the name argument should only contain 'project-name'
            instead of 'workspace-name/project-name'.
        username: Name of the user to remove from the project.
        workspace: Name of your Neptune workspace. If you specify it,
            change the format of the name argument to 'project-name' instead of 'workspace-name/project-name'.
            If None, it will be parsed from the name argument.
        api_token: Account's API token.
            If None, the value of the NEPTUNE_API_TOKEN environment variable is used.
            Note: To keep your token secure, use the NEPTUNE_API_TOKEN environment variable rather than placing your
            API token in plain text in your source code.

    Example:
        >>> from neptune import management
        >>> management.remove_project_member(
        ...     project="ml-team/classification",
        ...     username="johnny",
        ... )

    You may also want to check the management API reference:
    https://docs.neptune.ai/api/management
    """
    verify_type("project", project, str)
    verify_type("username", username, str)
    verify_type("workspace", workspace, (str, type(None)))
    verify_type("api_token", api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)
    project_identifier = normalize_project_name(name=project, workspace=workspace)

    params = {
        "projectIdentifier": project_identifier,
        "userId": username,
        **DEFAULT_REQUEST_KWARGS,
    }

    try:
        backend_client.api.deleteProjectMember(**params).response()
    except HTTPNotFound as e:
        raise ProjectNotFound(name=project_identifier) from e
    except HTTPUnprocessableEntity as e:
        raise UserNotExistsOrWithoutAccess(user=username, project=project_identifier) from e
    except HTTPForbidden as e:
        raise AccessRevokedOnMemberRemoval(user=username, project=project_identifier) from e


@with_api_exceptions_handler
def get_workspace_member_list(workspace: str, *, api_token: Optional[str] = None) -> Dict[str, str]:
    """Lists members of a Neptune workspace.

    Args:
        workspace: Name of the Neptune workspace.
        api_token: Account's API token.
            If None, the value of the NEPTUNE_API_TOKEN environment variable is used.
            Note: To keep your token secure, use the NEPTUNE_API_TOKEN environment variable rather than placing your
            API token in plain text in your source code.
    Returns:
        Dictionary with usernames as keys and workspace member roles ('admin', 'member') as values.

    Example:
        >>> from neptune import management
        >>> management.get_workspace_member_list(workspace="ml-team")

    You may also want to check the management API reference:
    https://docs.neptune.ai/api/management
    """
    verify_type("workspace", workspace, str)
    verify_type("api_token", api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)

    params = {"organizationIdentifier": workspace, **DEFAULT_REQUEST_KWARGS}

    try:
        result = backend_client.api.listOrganizationMembers(**params).response().result
        return {f"{m.registeredMemberInfo.username}": WorkspaceMemberRoleDTO.to_domain(m.role) for m in result}
    except HTTPNotFound as e:
        raise WorkspaceNotFound(workspace=workspace) from e


@with_api_exceptions_handler
def _get_raw_workspace_service_account_list(
    workspace_name: str, api_token: Optional[str] = None
) -> Dict[str, ServiceAccountDTO]:
    verify_type("workspace_name", workspace_name, str)
    verify_type("api_token", api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)

    params = {
        "organizationIdentifier": workspace_name,
        "deactivated": False,
        **DEFAULT_REQUEST_KWARGS,
    }

    try:
        result = backend_client.api.listServiceAccounts(**params).response().result
        return {f"{sa.displayName}": ServiceAccountDTO(name=sa.displayName, id=sa.id) for sa in result}
    except HTTPNotFound as e:
        raise WorkspaceNotFound(workspace=workspace_name) from e


@with_api_exceptions_handler
def get_workspace_service_account_list(workspace: str, *, api_token: Optional[str] = None) -> Dict[str, str]:
    """Lists service accounts of a Neptune workspace.

    Args:
        workspace: Name of the Neptune workspace.
        api_token: Account's API token.
            If None, the value of the NEPTUNE_API_TOKEN environment variable is used.
            Note: To keep your token secure, use the NEPTUNE_API_TOKEN environment variable rather than placing your
            API token in plain text in your source code.
    Returns:
        Dictionary with account names as keys and workspace member roles as values.
        Service accounts can only have the 'member' role in workspaces.

    Example:
        >>> from neptune import management
        >>> management.get_workspace_service_account_list(workspace="ml-team")

    You may also want to check the management API reference:
    https://docs.neptune.ai/api/management
    """
    service_accounts = _get_raw_workspace_service_account_list(workspace_name=workspace, api_token=api_token)

    return {
        service_account_name: WorkspaceMemberRoleDTO.to_domain("member")
        for service_account_name, _ in service_accounts.items()
    }


@with_api_exceptions_handler
def invite_to_workspace(
    *,
    username: Optional[str] = None,
    email: Optional[str] = None,
    workspace: str,
    api_token: Optional[str] = None,
    role: Union[WorkspaceMemberRole, str] = WorkspaceMemberRole.MEMBER,
    add_to_all_projects: bool = False,
) -> None:
    """Creates invitation to Neptune workspace.

    Args:
        username: username of the user to invite.
        email: email of the user to invite.
            Note: at least one of the above parameters are needed.
            If neither the username nor the email is passed, will raise ValueError.
            If both are filled, will raise ValueError.
        workspace: Name of your Neptune workspace.
        api_token: Account's API token.
            If None, the value of the NEPTUNE_API_TOKEN environment variable is used.
            Note: To keep your token secure, use the NEPTUNE_API_TOKEN environment variable rather than placing your
            API token in plain text in your source code.
        role: The workspace role that is to be granted to the invited user.
            You can choose between the following values: "admin" and "member".
        add_to_all_projects: Whether to add the user to all projects in the workspace.

    Example:
        >>> from neptune import management
        >>> management.invite_to_workspace(
        ...     username="user",
        ...     workspace="ml-team",
        ...     role="admin",
        ... )

    You may also want to check the management API reference:
    https://docs.neptune.ai/api/management
    """
    verify_type("workspace", workspace, str)
    verify_type("role", role, (WorkspaceMemberRole, str))
    verify_type("add_to_all_projects", add_to_all_projects, bool)
    verify_type("username", username, (str, type(None)))
    verify_type("email", email, (str, type(None)))
    verify_type("api_token", api_token, (str, type(None)))

    if username and email:
        raise ValueError("Cannot specify both `username` and `email`.")

    if username:
        invitee = username
        invitation_type = "user"
    elif email:
        invitee = email
        invitation_type = "emailRecipient"
    else:
        raise ValueError("Neither `username` nor `email` arguments filled. At least one needs to be passed")

    if isinstance(role, str):
        role = WorkspaceMemberRole(role)

    params = {
        "newOrganizationInvitations": {
            "invitationsEntries": [
                {
                    "invitee": invitee,
                    "invitationType": invitation_type,
                    "roleGrant": role.to_api(),
                    "addToAllProjects": add_to_all_projects,
                }
            ],
            "organizationIdentifier": workspace,
        },
        **DEFAULT_REQUEST_KWARGS,
    }

    backend_client = _get_backend_client(api_token=api_token)
    try:
        backend_client.api.createOrganizationInvitations(**params)
    except HTTPNotFound:
        raise WorkspaceOrUserNotFound(workspace=workspace, user=invitee)
    except HTTPConflict:
        raise UserAlreadyInvited(user=invitee, workspace=workspace)


@with_api_exceptions_handler
def get_project_service_account_list(
    project: str, *, workspace: Optional[str] = None, api_token: Optional[str] = None
) -> Dict[str, str]:
    """Lists service accounts assigned to a Neptune project.

    Args:
        project: The name of the project in Neptune in the form 'workspace-name/project-name'.
            If you pass the workspace argument, the name argument should only contain 'project-name'
            instead of 'workspace-name/project-name'.
        workspace: Name of your Neptune workspace. If you specify it,
            change the format of the name argument to 'project-name' instead of 'workspace-name/project-name'.
            If None, it will be parsed from the name argument.
        api_token: Account's API token.
            If None, the value of the NEPTUNE_API_TOKEN environment variable is used.
            Note: To keep your token secure, use the NEPTUNE_API_TOKEN environment variable rather than placing your
            API token in plain text in your source code.
    Returns:
        Dictionary with account names as keys and project member roles
        ('owner', 'contributor', 'viewer') as values.

    Example:
        >>> from neptune import management
        >>> management.get_project_service_account_list(
        ...     workspace="ml-team",
        ...     project="classification",
        ... )

    You may also want to check the management API reference:
    https://docs.neptune.ai/api/management
    """
    verify_type("project", project, str)
    verify_type("workspace", workspace, (str, type(None)))
    verify_type("api_token", api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)
    project_identifier = normalize_project_name(name=project, workspace=workspace)

    params = {"projectIdentifier": project_identifier, **DEFAULT_REQUEST_KWARGS}

    try:
        result = backend_client.api.listProjectServiceAccounts(**params).response().result
        return {f"{sa.serviceAccountInfo.displayName}": ProjectMemberRoleDTO.to_domain(sa.role) for sa in result}
    except HTTPNotFound as e:
        raise ProjectNotFound(name=project_identifier) from e


@with_api_exceptions_handler
def add_project_service_account(
    project: str,
    service_account_name: str,
    role: str,
    *,
    workspace: Optional[str] = None,
    api_token: Optional[str] = None,
):
    """Adds a service account to a Neptune project.

    Only project owners can add accounts as members.

    Args:
        project: The name of the project in Neptune in the form 'workspace-name/project-name'.
            If you pass the workspace argument, the name argument should only contain 'project-name'
            instead of 'workspace-name/project-name'.
        service_account_name: Name of the service account to add to the project.
        role: level of permissions the user or service account should have in a project.
            Can be set to:
             - 'viewer': can only view project content and members
             - 'contributor': can view and edit project content and only view members
             - 'owner': can view and edit project content and members
            For more information, see https://docs.neptune.ai/management/roles/
        workspace: Name of your Neptune workspace. If you specify it,
            change the format of the name argument to 'project-name' instead of 'workspace-name/project-name'.
            If None, it will be parsed from the name argument.
        api_token: Account's API token.
            If None, the value of the NEPTUNE_API_TOKEN environment variable is used.
            Note: To keep your token secure, use the NEPTUNE_API_TOKEN environment variable rather than placing your
            API token in plain text in your source code.

    Examples:
        >>> from neptune import management
        >>> management.add_project_service_account(
        ...     workspace="ml-team",
        ...     project="classification",
        ...     service_account_name="cicd@ml-team",
        ...     role="contributor",
        ... )

    You may also want to check the management API reference:
    https://docs.neptune.ai/api/management
    """
    verify_type("project", project, str)
    verify_type("service_account_name", service_account_name, str)
    verify_type("role", role, str)
    verify_type("workspace", workspace, (str, type(None)))
    verify_type("api_token", api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)
    workspace, project_name = extract_project_and_workspace(name=project, workspace=workspace)
    project_qualified_name = f"{workspace}/{project_name}"

    try:
        service_account = _get_raw_workspace_service_account_list(workspace_name=workspace, api_token=api_token)[
            service_account_name
        ]
    except KeyError as e:
        raise ServiceAccountNotFound(service_account_name=service_account_name, workspace=workspace) from e

    params = {
        "projectIdentifier": project_qualified_name,
        "account": {
            "serviceAccountId": service_account.id,
            "role": ProjectMemberRoleDTO.from_str(role).value,
        },
        **DEFAULT_REQUEST_KWARGS,
    }

    try:
        backend_client.api.addProjectServiceAccount(**params).response()
    except HTTPNotFound as e:
        raise ProjectNotFound(name=project_qualified_name) from e
    except HTTPConflict as e:
        service_accounts = get_project_service_account_list(project=project, workspace=workspace, api_token=api_token)
        service_account_role = service_accounts.get(service_account_name)

        raise ServiceAccountAlreadyHasAccess(
            service_account_name=service_account_name,
            project=project_qualified_name,
            role=service_account_role,
        ) from e


@with_api_exceptions_handler
def remove_project_service_account(
    project: str,
    service_account_name: str,
    *,
    workspace: Optional[str] = None,
    api_token: Optional[str] = None,
):
    """Removes a service account from a Neptune project.

    Only project owners can remove accounts.

    Args:
        project: The name of the project in Neptune in the form 'workspace-name/project-name'.
            If you pass the workspace argument, the name argument should only contain 'project-name'
            instead of 'workspace-name/project-name'.
        service_account_name: Name of the service account to remove from the project.
        workspace: Name of your Neptune workspace. If you specify it,
            change the format of the name argument to 'project-name' instead of 'workspace-name/project-name'.
            If None, it will be parsed from the name argument.
        api_token: Account's API token.
            If None, the value of the NEPTUNE_API_TOKEN environment variable is used.
            Note: To keep your token secure, use the NEPTUNE_API_TOKEN environment variable rather than placing your
            API token in plain text in your source code.

    Examples:
        >>> from neptune import management
        >>> management.remove_project_service_account(
        ...     workspace="ml-team",
        ...     project="classification",
        ...     service_account_name="cicd@ml-team",
        ... )

    You may also want to check the management API reference:
    https://docs.neptune.ai/api/management
    """
    verify_type("project", project, str)
    verify_type("service_account_name", service_account_name, str)
    verify_type("workspace", workspace, (str, type(None)))
    verify_type("api_token", api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)
    workspace, project_name = extract_project_and_workspace(name=project, workspace=workspace)
    project_qualified_name = f"{workspace}/{project_name}"

    try:
        service_account = _get_raw_workspace_service_account_list(workspace_name=workspace, api_token=api_token)[
            service_account_name
        ]
    except KeyError as e:
        raise ServiceAccountNotFound(service_account_name=service_account_name, workspace=workspace) from e

    params = {
        "projectIdentifier": project_qualified_name,
        "serviceAccountId": service_account.id,
        **DEFAULT_REQUEST_KWARGS,
    }

    try:
        backend_client.api.deleteProjectServiceAccount(**params).response()
    except HTTPNotFound as e:
        raise ProjectNotFound(name=project_qualified_name) from e
    except HTTPUnprocessableEntity as e:
        raise ServiceAccountNotExistsOrWithoutAccess(
            service_account_name=service_account_name, project=project_qualified_name
        ) from e
    except HTTPForbidden as e:
        raise AccessRevokedOnServiceAccountRemoval(
            service_account_name=service_account_name, project=project_qualified_name
        ) from e


def trash_objects(
    project: str,
    ids: Union[str, Iterable[str]],
    *,
    workspace: str = None,
    api_token: str = None,
) -> None:
    """Moves one or more Neptune objects to the project trash.

    Args:
        project: The name of the project in Neptune in the form 'workspace-name/project-name'.
            If you pass the workspace argument, the name argument should only contain 'project-name'
            instead of 'workspace-name/project-name'.
        ids: Neptune ID of object to trash (or list of multiple IDs).
            You can find the ID in the leftmost column of the table view, and in the "sys/id" field of each object.
        workspace: Name of your Neptune workspace. If you specify it,
            change the format of the name argument to 'project-name' instead of 'workspace-name/project-name'.
            If None, it will be parsed from the name argument.
        api_token: Account's API token.
            If None, the value of the NEPTUNE_API_TOKEN environment variable is used.
            Note: To keep your token secure, use the NEPTUNE_API_TOKEN environment variable rather than placing your
            API token in plain text in your source code.

    Examples:

        Trashing a run with the ID "CLS-1":
        >>> from neptune import management
        >>> management.trash_objects(project="ml-team/classification", ids="CLS-1")

        Trashing two runs and a model with the key "PRETRAINED":
        >>> management.trash_objects("ml-team/classification", ["CLS-2", "CLS-3", "CLS-PRETRAINED"])
        Note: Trashing a model object also trashes all of its versions.

    For more, see the docs: https://docs.neptune.ai/api/management/#trash_objects
    """
    verify_type("project", project, str)
    verify_type("workspace", workspace, (str, type(None)))
    verify_type("api_token", api_token, (str, type(None)))
    if ids is not None:
        if isinstance(ids, str):
            ids = [ids]
        else:
            verify_collection_type("ids", ids, str)

    leaderboard_client = _get_leaderboard_client(api_token=api_token)
    workspace, project_name = extract_project_and_workspace(name=project, workspace=workspace)
    project_qualified_name = f"{workspace}/{project_name}"

    qualified_name_ids = [QualifiedName(f"{workspace}/{project_name}/{container_id}") for container_id in ids]
    errors = list()
    for batch_ids in get_batches(qualified_name_ids, batch_size=TRASH_BATCH_SIZE):
        params = {
            "projectIdentifier": project_qualified_name,
            "experimentIdentifiers": batch_ids,
            **DEFAULT_REQUEST_KWARGS,
        }
        response = leaderboard_client.api.trashExperiments(**params).response()
        errors += response.result.errors

    for error in errors:
        logger.warning(error)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import threading
import time
from collections import namedtuple
from itertools import groupby
from queue import (
    Empty,
    Queue,
)

from bravado.exception import HTTPUnprocessableEntity

from neptune.legacy.exceptions import NeptuneException
from neptune.legacy.internal.channels.channels import (
    ChannelIdWithValues,
    ChannelNamespace,
    ChannelType,
    ChannelValue,
)
from neptune.legacy.internal.threads.neptune_thread import NeptuneThread

_logger = logging.getLogger(__name__)


class ChannelsValuesSender(object):
    _QUEUED_CHANNEL_VALUE = namedtuple(
        "QueuedChannelValue",
        [
            "channel_id",
            "channel_name",
            "channel_type",
            "channel_value",
            "channel_namespace",
        ],
    )

    __LOCK = threading.RLock()

    def __init__(self, experiment):
        self._experiment = experiment
        self._values_queue = None
        self._sending_thread = None
        self._user_channel_name_to_id_map = dict()
        self._system_channel_name_to_id_map = dict()

    def send(
        self,
        channel_name,
        channel_type,
        channel_value,
        channel_namespace=ChannelNamespace.USER,
    ):
        # Before taking the lock here, we need to check if the sending thread is not running yet.
        # Otherwise, the sending thread could call send() while being join()-ed, which would result
        # in a deadlock.
        if not self._is_running():
            with self.__LOCK:
                if not self._is_running():
                    self._start()

        if channel_namespace == ChannelNamespace.USER:
            namespaced_channel_map = self._user_channel_name_to_id_map
        else:
            namespaced_channel_map = self._system_channel_name_to_id_map

        if channel_name in namespaced_channel_map:
            channel_id = namespaced_channel_map[channel_name]
        else:
            response = self._experiment._create_channel(channel_name, channel_type, channel_namespace)
            channel_id = response.id
            namespaced_channel_map[channel_name] = channel_id

        self._values_queue.put(
            self._QUEUED_CHANNEL_VALUE(
                channel_id=channel_id,
                channel_name=channel_name,
                channel_type=channel_type,
                channel_value=channel_value,
                channel_namespace=channel_namespace,
            )
        )

    def join(self):
        with self.__LOCK:
            if self._is_running():
                self._sending_thread.interrupt()
                self._sending_thread.join()
                self._sending_thread = None
                self._values_queue = None

    def _is_running(self):
        return self._values_queue is not None and self._sending_thread is not None and self._sending_thread.is_alive()

    def _start(self):
        self._values_queue = Queue()
        self._sending_thread = ChannelsValuesSendingThread(self._experiment, self._values_queue)
        self._sending_thread.start()


class ChannelsValuesSendingThread(NeptuneThread):
    _SLEEP_TIME = 5
    _MAX_VALUES_BATCH_LENGTH = 100
    _MAX_IMAGE_VALUES_BATCH_SIZE = 10485760  # 10 MB

    def __init__(self, experiment, values_queue):
        super(ChannelsValuesSendingThread, self).__init__(is_daemon=False)
        self._values_queue = values_queue
        self._experiment = experiment
        self._sleep_time = self._SLEEP_TIME
        self._values_batch = []

    def run(self):
        while self.should_continue_running() or not self._values_queue.empty():
            try:
                sleep_start = time.time()
                self._values_batch.append(self._values_queue.get(timeout=max(self._sleep_time, 0)))
                self._values_queue.task_done()
                self._sleep_time -= time.time() - sleep_start
            except Empty:
                self._sleep_time = 0

            image_values_batch_size = sum(
                [
                    len(v.channel_value.y["image_value"]["data"] or [])
                    for v in self._values_batch
                    if v.channel_type == ChannelType.IMAGE.value
                ]
            )
            if (
                self._sleep_time <= 0
                or len(self._values_batch) >= self._MAX_VALUES_BATCH_LENGTH
                or image_values_batch_size >= self._MAX_IMAGE_VALUES_BATCH_SIZE
            ):
                self._process_batch()

        self._process_batch()

    def _process_batch(self):
        send_start = time.time()
        if self._values_batch:
            try:
                self._send_values(self._values_batch)
                self._values_batch = []
            except (NeptuneException, IOError):
                _logger.exception("Failed to send channel value.")
        self._sleep_time = self._SLEEP_TIME - (time.time() - send_start)

    def _send_values(self, queued_channels_values):
        def get_channel_metadata(value):
            return (
                value.channel_id,
                value.channel_name,
                value.channel_type,
                value.channel_namespace,
            )

        queued_grouped_by_channel = {
            channel_metadata: list(values)
            for channel_metadata, values in groupby(
                sorted(queued_channels_values, key=get_channel_metadata),
                get_channel_metadata,
            )
        }
        channels_with_values = []
        for channel_metadata in queued_grouped_by_channel:
            channel_values = []
            for queued_value in queued_grouped_by_channel[channel_metadata]:
                channel_values.append(
                    ChannelValue(
                        ts=queued_value.channel_value.ts,
                        x=queued_value.channel_value.x,
                        y=queued_value.channel_value.y,
                    )
                )
            channels_with_values.append(ChannelIdWithValues(*channel_metadata, channel_values))

        try:
            self._experiment._send_channels_values(channels_with_values)
        except HTTPUnprocessableEntity as e:
            message = "Maximum storage limit reached"
            try:
                message = e.response.json()["message"]
            finally:
                _logger.warning("Failed to send channel value: %s", message)
        except (NeptuneException, IOError):
            _logger.exception("Failed to send channel value.")



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "Artifact",
    "Boolean",
    "Datetime",
    "File",
    "Float",
    "GitRef",
    "Integer",
    "String",
    "FileSet",
    "StringSet",
    "FileSeries",
    "FloatSeries",
    "StringSeries",
]


from .atoms import (
    Artifact,
    Boolean,
    Datetime,
    File,
    Float,
    GitRef,
    Integer,
    String,
)
from .file_set import FileSet
from .series import (
    FileSeries,
    FloatSeries,
    StringSeries,
)
from .sets import StringSet
